import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import matplotlib.pyplot as plt

from numpy.lib.npyio import save
from torch.utils import data
from torchvision import transforms
from torch.utils.data import Dataset, DataLoader
from skimage import transform
from skimage import measure, img_as_ubyte
from skimage.morphology import binary_dilation, remove_small_objects, remove_small_holes
from skimage.measure import label, regionprops
from skimage.io import imread, imsave
from narsil.utils.transforms import  resizeOneImage, tensorizeOneImage
from narsil.utils.losses import WeightedUnetLoss, UnetLoss, WeightedUnetLossExact
from .datasets import mmDataMultiSpecies, phaseFolder, phaseTestDir
from .network import basicUnet, smallerUnet 

class trainNet(object):

	def __init__(self, dataDir, species, transforms, modelParameters, optimizationParameters,
					validationTransforms=None, validation=True):
		self.dataDir = dataDir
		self.species = species
		self.transforms = transforms
		self.validation = validation
		self.validationTransforms = validationTransforms

		self.modelParameters = modelParameters
		self.optimizationParameters = optimizationParameters
		

		
		# device initialization
		self.device = torch.device(self.modelParameters['device'] if torch.cuda.is_available() else "cpu")

		# initalize dataset
		self.initializeDataset()

		# initialize net
		self.initializeNet()

		# initialize optimizer
		self.initializeOptimizer()

		# initialize loss function
		self.initializeLossFunction()

		self.sendToDevice()


	def initializeDataset(self):
		self.trainDataset = mmDataMultiSpecies(self.dataDir, self.species, self.transforms, datasetType='train',
									includeWeights=self.modelParameters['includeWeights']) 
		# on validationDataset, validate on the full image, instead of the crops
		self.validationDataset = mmDataMultiSpecies(self.dataDir, self.species, transforms=self.validationTransforms,
								 datasetType='validation', includeWeights=self.modelParameters['includeWeights'])

		# dataLoaders initialization
		self.trainDataLoader = DataLoader(self.trainDataset, batch_size = self.optimizationParameters['batchSize'],
									shuffle=True, num_workers = self.optimizationParameters['cores'])
		print("Training DataLoaders initialized .... ")
		self.validationDataLoader = DataLoader(self.validationDataset, batch_size=1, 
									shuffle=False, num_workers=self.optimizationParameters['cores'])
		print("Validation DataLoaders initialized .... ")

		

	def initializeNet(self):
		if self.modelParameters['netType'] == 'big':
			self.net = basicUnet(self.modelParameters['transposeConv'])
		elif self.modelParameters['netType'] == 'small':
			self.net = smallerUnet(self.modelParameters['transposeConv'])
		
		print("Network initalized successfully ...")

	def initializeLossFunction(self):
		#self.lossFunction = nn.BCELoss()
		#self.lossFunction = WeightedUnetLoss()
		self.lossFunction = WeightedUnetLossExact()

		print("Loss function initialized successfully ...")

	def initializeOptimizer(self):
		self.optimizer = optim.Adam(self.net.parameters(),lr = self.optimizationParameters['learningRate'])
		self.scheduler = optim.lr_scheduler.StepLR(self.optimizer,
							step_size = self.optimizationParameters['schedulerStep'],
							gamma = self.optimizationParameters['schedulerGamma'] )
		
		print("Optimizer and scheduler initialized successfully ... ")


	def sendToDevice(self):
		self.net.to(self.device)

	def train(self):
		print("Training started .... ")
		self.losses_train = []
		self.losses_validation = []
		for epoch in range(self.optimizationParameters['nEpochs']):
			epoch_train_loss = self.trainEpoch(epoch)
			print(f"Training Epoch {epoch + 1} done --- {epoch_train_loss}")
			self.losses_train.append(epoch_train_loss)

			# validation
			if self.validation == True:
				epoch_validation_loss = self.validateEpoch(epoch)
				self.losses_validation.append(epoch_validation_loss)

			self.scheduler.step()

	def trainEpoch(self, epoch):
		epochLoss = 0.0
		for i_batch, data in enumerate(self.trainDataLoader, 0):

			phase, mask, weights= data['phase'].to(self.device), data['mask'].to(self.device), data['weights'].to(self.device)

			self.optimizer.zero_grad()

			mask_pred = self.net(phase)
			loss = self.lossFunction(mask_pred, mask, weights)

			epochLoss += loss.item()
			print(f"Epoch : {epoch + 1}, ImageBatch: {i_batch} -- loss: {loss.item()}")

			loss.backward()
			self.optimizer.step()
		
		# run the model through a validation loop
		return epochLoss/i_batch

	def validateEpoch(self, epoch):
		self.net.eval()
		validation_epoch_loss = 0.0
		with torch.no_grad():
			for i_batch, data in enumerate(self.validationDataLoader, 0):
				phase, mask, weights = data['phase'].to(self.device), data['mask'].to(self.device), data['weights'].to(self.device)
				mask_pred = self.net(phase)
				loss_per_image = self.lossFunction(mask_pred, mask, weights)
				validation_epoch_loss += loss_per_image.item()
		# reset net back to train mode after evaluating
		self.net.train()
		print(f"Validation run of epoch: {epoch + 1} done ...")
		return validation_epoch_loss/i_batch 

	def plotLosses(self, ylim=[0.1, 0.4]):
		plt.figure()
		epochs = range(1, self.optimizationParameters['nEpochs'] + 1)
		plt.plot(epochs, self.losses_train, label='Train')
		plt.plot(epochs, self.losses_validation, label='Validation')
		plt.legend()
		plt.xlabel('Epochs')
		plt.ylim(ylim)
		plt.ylabel("Dice loss + 0.5 * Weighted BCE loss")
		plt.title("Covergence of the network")
		plt.show()
		

	def plotData(self, idx):

		fig, (ax1, ax2, ax3) = plt.subplots(nrows=1, ncols=3)
		datapoint = self.trainDataset[idx]
		ax1.imshow(datapoint['phase'].numpy().squeeze())
		ax2.imshow(datapoint['mask'].numpy().squeeze())
		ax3.imshow(datapoint['weights'].numpy().squeeze())
		ax1.set_xlabel(datapoint['phaseFilename'])
		ax2.set_xlabel(datapoint['maskFilename'])
		ax3.set_xlabel(datapoint['weightsFilename'])
		plt.show()


	def save(self, path):
		transformsUsed = str(self.transforms)
		
		savedModel = {
			'modelParameters' : self.modelParameters,
			'optimizationParameters': self.optimizationParameters,
			'trasnformsUsed' : transformsUsed,
			'species' : self.species,
			'model_state_dict': self.net.state_dict(),
		}
		torch.save(savedModel,path)



def testNet(modelPath, phaseDir, saveDir, transforms,
	threshold = None, device = 'cuda:1', fileformat ='.tiff', 
	plotContours='False', contoursThreshold=0.9, removeSmallObjects=False,
	addNoise =False):
	
	savedModel = torch.load(modelPath)
	if savedModel['modelParameters']['netType'] == 'big':
		net = basicUnet(savedModel['modelParameters']['transposeConv'])
	elif savedModel['modelParameters']['netType'] == 'small':
		net = smallerUnet(savedModel['modelParameters']['transposeConv'])

	device = torch.device(device if torch.cuda.is_available() else "cpu")

	#resizing = resizeOneImage((1040, 2048), (1040, 2048))
	#resizing = resizeOneImage((2048, 4096), (1024, 2048))
	#tensorizing = tensorizeOneImage(1)

	#transform = transforms.Compose([resizing, tensorizing])

	motherMachineDataset = phaseTestDir(phaseDir, transform=transforms, phase_fileformat=fileformat,
				addNoise=addNoise, flip=False)

	motherMachineDataLoader = DataLoader(motherMachineDataset, batch_size=1, shuffle=True, num_workers=6)

	net.load_state_dict(savedModel['model_state_dict'])
	net.to(device)
	net.eval()

	numTest = len(motherMachineDataset)
	print(f"Length of dataset is {numTest}")

	# test loop
	with torch.no_grad():
		for i_batch, data in enumerate(motherMachineDataLoader, 0):

			# get the image tensors
			phase = data.to(device)
			# zero the paramter gradients
			# forward + backward + optimize
			mask_pred = net(phase) 

			# set sigmoid if the net gives 
			mask_pred = torch.sigmoid(mask_pred)

			#mask_pred = torch.sigmoid(mask_pred)
			if threshold != None:
				mask_pred = mask_pred.to("cpu").numpy().squeeze(0).squeeze(0) >= threshold
				if removeSmallObjects == True:
					mask_pred_labeled = label(mask_pred)
					mask_cleaned = remove_small_objects(mask_pred_labeled, min_size=40)
					mask_cleaned = remove_small_holes(mask_cleaned > 0, area_threshold=30)
					mask_pred = mask_cleaned	
			else:
				mask_pred = mask_pred.to("cpu").numpy().squeeze(0).squeeze(0) 

			phase_save = phase.to("cpu").numpy().squeeze(0).squeeze(0)

			if plotContours == False:
				plt.figure()
				plt.imshow(mask_pred)
				plt.show(block=False)

			if plotContours == True:
				contours = measure.find_contours(mask_pred, contoursThreshold)
				fig, ax = plt.subplots()
				ax.imshow(phase_save, cmap= plt.cm.gray)

				for contour in contours:
					ax.plot(contour[:, 1], contour[:, 0], linewidth=1, color='r')
				plt.show(block=False)
#        print(f"ImageBatch: {i_batch} -- loss: {loss.item()}")
			if saveDir != None:
				imsave(saveDir + str(i_batch) + '.tif', img_as_ubyte(mask_pred), plugin='tifffile', compress = 6, check_contrast=False)
	#    print(f"Epoch {epoch+1} Done --- Loss per Image: {epoch_loss/numTrain}")


			if i_batch == 10:
				break

			

	

	





import wx
import numpy as np
import pickle
from narsil.tracking.siamese.trainDatasets import oneSet

class mainWindow(wx.Frame):
    """ Main windows that helps loading data from one set of tracks 
    and annotate them
    """
    def __init__(self, parent, title):
        super(mainWindow, self).__init__(parent, title=title, size=(1000, 1000))
        self.InitUI()
        self.pathname = ''
        self.oneset = None
        self.idx1 = 0
        self.idx2 = 1
        self.blobState = "normal"
        self.links = {}
        self.leftBbox = -1
        self.rightBbox = -1


    def InitUI(self):

        # Menubar stuff
        menubar = wx.MenuBar()
        fileMenu = wx.Menu()
        fileItem = fileMenu.Append(wx.ID_EXIT, 'Quit', 'Quit Application')
        menubar.Append(fileMenu, '&File')
        self.SetMenuBar(menubar)
        self.Bind(wx.EVT_MENU, self.onQuit, fileItem)

        # Buttons to load, scroll, change, save, etc
        self.panel = wx.Panel(self)
        self.panel.SetBackgroundColour('#FFFFFF')
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        btnLoad = wx.Button(self.panel, label='Load', size=(150, 30))
        hbox.Add(btnLoad, flag=wx.LEFT|wx.TOP, border = 5)
        btnNext = wx.Button(self.panel, label='Next', size=(150, 30) )
        hbox.Add(btnNext, flag = wx.LEFT|wx.TOP, border = 5)
        btnReloadCurrent = wx.Button(self.panel, label='Relaod Current Step', size=(150, 30))
        hbox.Add(btnReloadCurrent, flag= wx.LEFT| wx.TOP, border=5)
        btnClose = wx.Button(self.panel, label='Close', size=(150, 30))
        hbox.Add(btnClose, flag = wx.LEFT | wx.TOP, border = 5)
        btnSave = wx.Button(self.panel, label = 'Save', size=(150, 30))
        hbox.Add(btnSave, flag = wx.LEFT | wx.TOP, border = 5)
        btnPrintMatrix = wx.Button(self.panel, label = 'Print Matrix', size=(150, 30))
        hbox.Add(btnPrintMatrix, flag = wx.LEFT | wx.TOP, border = 5)
        hbox1 = wx.BoxSizer(wx.HORIZONTAL)
        btnShowBoxes = wx.Button(self.panel, label = 'Show Boxes', size=(150, 30))
        hbox1.Add(btnShowBoxes, flag = wx.LEFT | wx.TOP, border = 5)

        vbox_buttons = wx.BoxSizer(wx.VERTICAL)
        #Normal segmentation
        self.normal = wx.RadioButton(self.panel, label = 'Normal', pos=(20, 300), style=wx.RB_GROUP)
        self.normal.Bind(wx.EVT_RADIOBUTTON, self.setBlobState)
        self.badSeg = wx.RadioButton(self.panel, label = 'Bad Segmentation', pos=(20, 320))
        self.badSeg.Bind(wx.EVT_RADIOBUTTON, self.setBlobState)
        self.notCell = wx.RadioButton(self.panel, label = 'Not Cell', pos=(20, 340))
        self.notCell.Bind(wx.EVT_RADIOBUTTON, self.setBlobState)
        self.ignoreCell = wx.RadioButton(self.panel, label = 'Ignore Cell', pos = (20, 360))
        self.ignoreCell.Bind(wx.EVT_RADIOBUTTON, self.setBlobState)
        vbox_buttons.Add(self.normal)
        vbox_buttons.Add(self.badSeg)
        vbox_buttons.Add(self.notCell)
        vbox_buttons.Add(self.ignoreCell)

        #button bindings
        self.Bind(wx.EVT_BUTTON, self.load,id = btnLoad.GetId())
        self.Bind(wx.EVT_BUTTON, self.next,id = btnNext.GetId())
        self.Bind(wx.EVT_BUTTON, self.reload,id = btnReloadCurrent.GetId())
        self.Bind(wx.EVT_BUTTON, self.close, id = btnClose.GetId())
        self.Bind(wx.EVT_BUTTON, self.save, id = btnSave.GetId())
        self.Bind(wx.EVT_BUTTON, self.printMatrix, id = btnPrintMatrix.GetId())
        

        # show dummy images
        self.firstImage = wx.StaticBitmap(self.panel, wx.ID_ANY, wx.Bitmap("gui/1.jpeg", wx.BITMAP_TYPE_ANY))
        self.secondImage = wx.StaticBitmap(self.panel, wx.ID_ANY, wx.Bitmap("gui/2.png", wx.BITMAP_TYPE_ANY))
        self.firstImage.SetPosition((300, 50))
        self.secondImage.SetPosition((500, 50))
        self.firstImage.Bind(wx.EVT_LEFT_DOWN, self.onLeftDown)
        self.secondImage.Bind(wx.EVT_LEFT_UP, self.onLeftUp)
        self.currentLine = [0, 0, 0 , 0]

        # set panel sizer for the button hbox
        vbox = wx.BoxSizer(wx.VERTICAL)
        vbox.Add(hbox)
        vbox.Add(hbox1)
        vbox.Add(vbox_buttons)
        self.panel.SetSizer(vbox)

    def setBlobState(self, e):
        normal_set = self.normal.GetValue()
        badSeg_set = self.badSeg.GetValue()
        notCell_set = self.notCell.GetValue()
        ignoreCell_set = self.ignoreCell.GetValue()
        if normal_set == True:
            print("Normal cell")
            self.blobState = "normal"
        elif badSeg_set == True:
            print("Bad segmentation")
            self.blobState = "badSeg"
        elif notCell_set == True:
            print("Not cell set")
            self.blobState = "notCell"
        elif ignoreCell_set == True:
            print("Ignore cell set")
            self.blobState = "ignoreCell"



    def loadImages(self, idx1, idx2):
        self.img1 = wx.Image(self.oneset[idx1]['filename'], wx.BITMAP_TYPE_TIFF).ConvertToBitmap()
        self.img2 = wx.Image(self.oneset[idx2]['filename'], wx.BITMAP_TYPE_TIFF).ConvertToBitmap()
        self.firstImage.SetBitmap(wx.Bitmap(self.img1))
        self.secondImage.SetBitmap(wx.Bitmap(self.img2))
        self.dict_index = str(idx1) + "_" + str(idx2)
        self.links[self.dict_index] = np.zeros((len(self.oneset[idx1]['bbox']), len(self.oneset[idx2]['bbox'])))
        self.local_links_OF = np.array([], dtype = int)



    def findObjectNumber(self, idx, x, y):
        # This function returns the object number on the
        bbox_list = self.oneset[idx]['bbox']
        for i in range(len(bbox_list)):
            if y >= bbox_list[i][0] and y < bbox_list[i][2]:
                return (i, bbox_list[i][2] - bbox_list[i][0], bbox_list[i][3] - bbox_list[i][1], bbox_list[i][1], bbox_list[i][0])


    def onLeftDown(self, e):
        x, y = e.GetPosition()
        self.currentLine[0] = x + 300
        self.currentLine[1] = y + 50
        # find the bbox here
        bbox_number, height, width, draw_x, draw_y  = self.findObjectNumber(self.idx1, x, y)
        print("Bbox1 : ", bbox_number)
        self.leftBbox = bbox_number
        # draw the rectangle of the bbox
        self.dc2 = wx.ClientDC(self)
        if self.blobState == 'normal':
            self.dc2.SetPen(wx.Pen('#c56c00', 1, wx.SOLID))
        else:
            self.dc2.SetPen(wx.Pen('#004fc5', 1, wx.SOLID))
        self.dc2.DrawRectangle(draw_x + 700, draw_y + 50, width, height)

    
    def onLeftUp(self, e):
        x, y = e.GetPosition()
        self.currentLine[2] = x + 500
        self.currentLine[3] = y + 50
        self.drawCurrentLine()
        # find the bbox here
        bbox_number, height, width, draw_x, draw_y = self.findObjectNumber(self.idx2, x, y) 
        self.rightBbox = bbox_number
        print("Bbox2 : ", bbox_number)
        self.dc3 = wx.ClientDC(self)
        if self.blobState == 'normal':
            self.dc3.SetPen(wx.Pen('#c56c00', 1, wx.DOT))
        else:
            self.dc3.SetPen(wx.Pen('#004fc5', 1, wx.DOT))
        self.dc3.DrawRectangle(draw_x + 900, draw_y + 50, width, height)

        # After drawing the boxes, now set the links
        self.links[self.dict_index][self.leftBbox, self.rightBbox] = 1.0
        self.draw_links_OF = np.array([[self.idx1, self.leftBbox, self.rightBbox]])
        if self.local_links_OF.size == 0:
            self.local_links_OF = self.draw_links_OF
        else:
            self.local_links_OF = np.concatenate((self.local_links_OF, self.draw_links_OF))
        


    def printMatrix(self, e):
        print("----- Local links Matrix format ----- ")
        print(self.links[self.dict_index])
        print("----- Pipeline format: Local Links ------------")
        print(self.local_links_OF)
        print("----- Pipeline format: Globl links ------------")
        print(self.blobs_links_OF)




    def drawCurrentLine(self):
        self.dc = wx.ClientDC(self)
        self.dc.SetPen(wx.Pen('#4c4c4c', 1, wx.SOLID))
        self.dc.DrawLine(*self.currentLine)

    
    def save(self, e):
        savingFilename = self.pathname + 'links.npy'
        np.save(savingFilename, self.links)
        savingJustLinksFilename = self.pathname + 'justlinks'
        np.save(savingJustLinksFilename, self.blobs_links_OF )
        print(savingFilename, " SAVED !!!!")
        print(savingJustLinksFilename, " SAVED !!!!!")

    def load(self, e):
        # This function opens a directory dialog prompt to pick the directory containing the image sequence
        dirdialog = wx.DirDialog(self, message = "Open blob sequence directory", defaultPath='/home/batnet/Documents/trackingdata/blobsSingle/', style=wx.DD_DEFAULT_STYLE)
        if dirdialog.ShowModal() == wx.ID_CANCEL:
            return
        self.pathname = dirdialog.GetPath() + '/'
        print("Path Set:" , self.pathname)
        self.oneset = oneSet(self.pathname)
        #print("Oneset initialized:", self.oneset[0])
        print(self.oneset)
        self.idx1 = 0
        self.idx2 = 1
        self.loadImages(self.idx1, self.idx2)
        self.blobs_links_OF = np.array([], dtype = int)

    def next(self, e):
        # if not last time point increment indices
        if self.idx2 == len(self.oneset) - 1:
            wx.MessageBox('Sequence completed', 'Info', wx.OK | wx.ICON_STOP)
            if self.blobs_links_OF.size == 0:
                self.blobs_links_OF = self.local_links_OF
            else:
                self.blobs_links_OF = np.concatenate((self.blobs_links_OF, self.local_links_OF))
 
            return
        else:
            self.idx1 += 1
            self.idx2 +=1
            self.dc.Clear()
            if self.blobs_links_OF.size == 0:
                self.blobs_links_OF = self.local_links_OF
            else:
                self.blobs_links_OF = np.concatenate((self.blobs_links_OF, self.local_links_OF))
        self.loadImages(self.idx1, self.idx2)



    def reload(self, e):
        if self.pathname == None:
            return
        else:
            #self.idx1 = 0
            #self.idx2 = 1
            self.loadImages(self.idx1, self.idx2)
            self.dc.Clear()
            
    
    def close(self, e):
        self.Close()


    def onQuit(self, e):
        self.Close()


def run():
    app = wx.App()
    gui = mainWindow(None, title="Tracking train dataset creation")
    gui.Show()
    app.MainLoop()

#if __name__ == '__main__':
#    main()


import glob
import os
import random
from skimage import io
from skimage.exposure.exposure import adjust_gamma, equalize_hist
from skimage.exposure.histogram_matching import match_histograms
import torch
import numpy as np
from torch.utils.data import Dataset, DataLoader
from matplotlib.widgets import Slider, Button, CheckButtons, RadioButtons
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import pathlib
from narsil.utils.transforms import resizeOneImage, tensorizeOneImage, stripAxis
from torchvision import transforms
from skimage.transform import resize, rotate


class mmDataset(Dataset):
    """ mm Segmentation Mask dataset for Training/Testing/Validation"""

    def __init__(self, phase_dir, mask_dir, scaleFactor=None,transform = None,
             phase_fileformat='.tiff', mask_fileformat = '.tiff', add_noise = False,
             augmentedIllumination=False):
        """
        Args:
              phase_dir (string): Name of the phase channel directory
              mask_dir (string): Name of the mask directory (use ImageJ)
              phase_transform : A series of transforms you want to apply to the images
                                (usually a crop and rotate if needed)
              mask_transform: A series of transforms that is needed to generate a segmask
                                (or just some crops, same as phase, if they are already
                                 preprocessed in ImageJ)   
              fileformat: Give the fileformat like .tiff, or .png for 
        """
        # read the folder and count number of images and set the transform
        self.phase_fileformat = phase_fileformat
        self.mask_fileformat = mask_fileformat
        self.n_images = len(glob.glob(phase_dir + "*" + phase_fileformat))
        self.phase_dir = phase_dir
        self.mask_dir = mask_dir
        self.indices = [filename.split('.')[0].split('/')[-1] for filename in glob.glob(phase_dir + "*" + phase_fileformat)] 
        self.transform = transform
        self.add_noise = add_noise
        self.scaleFactor = scaleFactor
        self.augmentedIllumination = augmentedIllumination

    def __len__(self):
        return self.n_images

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        phase_img_name = self.phase_dir + self.indices[idx] + self.phase_fileformat
        #mask_img_name = self.mask_dir + self.indices[idx] + '_mask' + self.mask_fileformat
        mask_img_name = self.mask_dir + self.indices[idx] + self.mask_fileformat
        
        phase_img = io.imread(phase_img_name, as_gray=True)
        #if random.random() > 0.25:
        #phase_img = adjust_gamma(phase_img, gamma=1.2)

        phase_img = phase_img.astype('float32')
        #phase_img = equalize_hist(phase_img)
        mask_img = io.imread(mask_img_name, as_gray=True).astype('bool')

        #height, width = phase_img.shape
        #if self.augmentedIllumination and random.random() > 0.5:
        #    phase_img = phase_img + np.array([np.linspace(1000, 8000, num=width),]*height)

        #phase_original = np.copy(phase_img)
        #mask_original = np.copy(mask_img)

        #phase_mean  = np.mean(phase_img)
        #phase_img = phase_img
        #phase_std = np.std(phase_img)

        #phase_img = (phase_img - phase_mean) / phase_std
        
        # zero everything that is above 2 standard deviations
        
        #indices = np.where(phase_img > 2)
        #mask_img[indices[0], indices[1]] = 0
        #if self.add_noise:
        #    rand_num = np.random.normal(0,0.15, phase_img.shape)
        #    phase_img = phase_img + rand_num

        sample = {'phase': phase_img, 'mask': mask_img , 'phase_filename': phase_img_name, 'mask_filename': mask_img_name}

        if self.transform:
            sample = self.transform(sample)

        return sample

class phaseFolder(Dataset):
    """
    Binding class that brings phase images in a directory together for batching and segmenting

    main Directory is going to be a positon directory and not the main main Dir containing positions
    """

    def __init__(self, mainDirectory, phaseDirName = "phase/", transform = None, phase_fileformat = '.tiff', addNoise = False, flip = False):
        self.mainDirectory = mainDirectory
        self.phase_fileformat = phase_fileformat
        self.transform = transform
        self.indices = [filename.split('.')[0].split('/')[-1] for filename in glob.glob(mainDirectory + phaseDirName +  "*" + phase_fileformat)]
        self.indices.sort()
        self.n_images = len(self.indices)
        self.phaseDirName = phaseDirName
        self.addNoise = addNoise
        self.flip = flip



    def __len__(self):
        return self.n_images

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        phase_img_name = self.mainDirectory + self.phaseDirName + self.indices[idx] + self.phase_fileformat

        phase_img = io.imread(phase_img_name, as_gray = True).astype('float32')
        #phase_img = match_histograms(phase_img, self.img_ref).astype('float32')

        phase_mean = np.mean(phase_img)
        phase_std = np.std(phase_img)

        phase_img_normalized = (phase_img - phase_mean) / phase_std


        if self.addNoise:
            rand_num = np.random.normal(0, 0.15, phase_img.shape)
            phase_img_normalized = phase_img_normalized + rand_num

        if self.flip:
            phase_img_normalized = rotate(phase_img_normalized, angle = 180)
            phase_img = rotate(phase_img, angle=180)
        

        if self.transform:
            phase_img_normalized = self.transform(phase_img_normalized)
            phase_img = self.transform(phase_img)

        return {'phase' : phase_img_normalized, 'phase_filename' : phase_img_name, 'phase_original': phase_img}

class phaseTestDir(object):
    """
    Binding class that brings phase images in a directory together for batching and segmenting

    main Directory is going to be a positon directory and not the main main Dir containing positions
    """
    def __init__(self, phaseDirectory, transform = None, phase_fileformat = '.tiff', addNoise = False, flip = False):
        self.phaseDirectory = phaseDirectory
        self.phase_fileformat = phase_fileformat
        self.transform = transform
        self.indices = [filename.split('.')[0].split('/')[-1] for filename in glob.glob(phaseDirectory + "*" + phase_fileformat)]
        self.n_images = len(self.indices)
        self.addNoise = addNoise
        self.flip = flip


    def __len__(self):
        return self.n_images

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        phase_img_name = self.phaseDirectory + self.indices[idx] + self.phase_fileformat

        phase_img = io.imread(phase_img_name, as_gray = True)
        #phase_img = adjust_gamma(phase_img, gamma=1.2)
        phase_img = phase_img.astype('float32')

        


        phase_mean = np.mean(phase_img)
        phase_std = np.std(phase_img)

        phase_img_normalized = (phase_img - phase_mean) / phase_std

        #phase_img[phase_img > 1.3] += 0.5

        #phase_img = match_histograms(phase_img, self.imgref)
        #phase_img = equalize_hist(phase_img)

        #phase_img[phase_img > 1.0] += 0.5
        

        if self.addNoise:
            rand_num = np.random.normal(0, 0.15, phase_img_normalized.shape)
            phase_img_normalized = phase_img_normalized + rand_num

        if self.flip:
            phase_img_normalized = rotate(phase_img_normalized, angle = 180)
        

        if self.transform:
            phase_img_normalized = self.transform(phase_img_normalized)

        return phase_img_normalized


"""
Class for clikcing through the segmentation and phase images results of any directory
using a GUI to verify quality of the network visually.
"""
class segViewer(object):

    def __init__(self, phaseDir, net, threshold, addNoise = False, shrinkInHalf = False, device = "cpu",fileformat='.tiff'):
        self.phaseDir = phaseDir
        self.fileformat = fileformat
        self.threshold = threshold
        self.net = net # net should already be loaded and in eval state
        self.transforms = None
        self.shrinkInHalf = shrinkInHalf
        self.addNoise = addNoise
        self.strip = stripAxis()
        self.device = torch.device(device)
        self.filenames = glob.glob(self.phaseDir + "*" + self.fileformat)

        self.fig, self.ax = plt.subplots(1, 1, num=str(self.phaseDir))
        self.axcolor = 'lightgoldenrodyellow'
        plt.subplots_adjust(left = 0.25,bottom=0.25)


        # set the currnetPhase Image and segImage
        self.currentImageIndex = 0
        self.currentPhaseImage = self.__getitem__(self.currentImageIndex) # will be a tensor
        with torch.no_grad():
            self.currentSegImage = self.net(self.currentPhaseImage) > self.threshold

        self.pltImage = self.ax.imshow(self.strip(self.currentPhaseImage))

        # buttons
        self.nextax = plt.axes([0.65, 0.025, 0.1, 0.03])
        self.previousax = plt.axes([0.8, 0.025, 0.1, 0.03])
        self.nextButton = Button(self.nextax, 'Next', color=self.axcolor, hovercolor='0.975')
        self.previousButton = Button(self.previousax, 'Previous', color=self.axcolor, hovercolor='0.975')
        self.nextButton.on_clicked(self.next)
        self.previousButton.on_clicked(self.previous)

        # toggle between phase and seg
        self.toggleNames = ('Phase', 'Segmentation')
        self.togglesax = plt.axes([0.025, 0.5, 0.15, 0.15])
        self.toggles = RadioButtons(self.togglesax, self.toggleNames)
        self.toggles.on_clicked(self.updateImage)

        plt.pause(0.1)
    
    # this will give out the tensor needed to go directly into the net
    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        phaseFileName = self.filenames[idx]
        phaseImage = io.imread(phaseFileName).astype('float32')
        
        (height, width) = phaseImage.shape
        if height%32 != 0:
            reshapeHeight = (height//32 + 1) * 32
        else:
            reshapeHeight = height
        
        if width%32 != 0:
            reshapeWidth = (width//32 + 1) * 32
        else:
            reshapeWidth = width

        resizeShape = (reshapeHeight, reshapeWidth)

        phase_mean = np.mean(phaseImage)
        phase_std = np.std(phaseImage)
        phaseImage = (phaseImage - phase_mean) / phase_std

        if self.addNoise:
            rand_num = np.random.normal(0, 0.15, phaseImage.shape)
            phaseImage = phaseImage + rand_num


        if self.shrinkInHalf:
            shrinkedHeight = reshapeHeight//2
            shrinkedWidth = reshapeWidth//2
            shrinkShape = (shrinkedHeight, shrinkedWidth)
            self.transforms = transforms.Compose([resizeOneImage(resizeShape,shrinkShape), tensorizeOneImage(2)])
        else:
            self.transforms = transforms.Compose([resizeOneImage(resizeShape, resizeShape),tensorizeOneImage(2)])

        if self.transforms != None:
            phaseImage = self.transforms(phaseImage)
        
        print(phaseImage.shape)

        return phaseImage.to(self.device)


    def __len__(self):
        return len(self.filenames)

    def next(self, val):
        if self.currentImageIndex != self.__len__() - 1:
            self.currentImageIndex += 1
        
        self.currentPhaseImage = self.__getitem__(self.currentImageIndex)
        with torch.no_grad():
            self.currentSegImage = self.net(self.currentPhaseImage) > self.threshold

        self.redraw()    

    def redraw(self):
        if self.toggles.value_selected == 'Phase':
            print("Phase is active")
            self.pltImage.set_data(self.strip(self.currentPhaseImage))
            self.fig.canvas.draw()
        elif self.toggles.value_selected == 'Segmentation':
            self.pltImage.set_data(self.strip(self.currentSegImage))
            self.fig.canvas.draw()


    def previous(self, val):
        if self.currentImageIndex != 0:
            self.currentImageIndex -= 1

        self.currentPhaseImage = self.__getitem__(self.currentImageIndex)
        with torch.no_grad():
            self.currentSegImage = self.net(self.currentPhaseImage) > self.threshold

        self.redraw()

    def updateImage(self, label):
        if label == 'Segmentation':
            self.pltImage.set_data(self.strip(self.currentSegImage))
            print("Segmenation",self.currentSegImage.shape)
            self.fig.canvas.draw()
        elif label == 'Phase':
            self.pltImage.set_data(self.strip(self.currentPhaseImage))
            print("Phase", self.currentPhaseImage.shape)
            self.fig.canvas.draw()





class fishdataFolder(object):
    """
    Binding class that brings information about 
    a particular channel of the flourescent channels
    Initialization:
    Parameters:
        1. fishMainDirectory (string) ---> Main directory where the Positions Folders are
        2. channel names (list of strings): Ex --> 488, cy5, cy3, etc 
        3. species -> channel map (dictionary) : Ex --> {'ecoli':[488, cy5], 'klebsiella': [cy3], 'cocci': [488]}
        4. flipPositions -> range Object or a list of position that need to be flipped
    """

    def __init__(self, fishPositionDir, channelNames, speciesChannelMap, flipPositions=None):

        self.fishPositionDir = fishPositionDir
        self.channelNames = channelNames
        self.speciesChannelMap = speciesChannelMap
        self.flipPositions = ['Pos' + str(i) for i in flipPositions]
        self.positionNumber = int(self.fishPositionDir.split('/')[-2][3:])
        # Grab the position names for the main directory
        #self.positions = [positionName for positionName in os.listdir(self.mainDirectory) if os.path.isdir(os.path.join(self.mainDirectory, positionName))]
        # sort positions based on its numbers
        #self.positions = sorted(self.positions, key= lambda position: int(position[3:]))

    def __len__(self):
        return len(self.channelNames)

    def __getitem__(self, idx):
        
        channelDirectories = {}
        for channelName in self.channelNames:
            channelDirectories[channelName] = self.mainDirectory + position + '/' + channelName + '/'

        channelDirectories['position'] = position
        if position in self.flipPositions:
            channelDirectories['flip'] = True
        else:
            channelDirectories['flip'] = False
        return channelDirectories


    def printSpeciesProbeMap(self):
        return None


class mmDataMultiSpecies(object):

    def __init__(self, trainingDataDir, species, transforms = None, datasetType='train', includeWeights=False):
        self.trainingDataDir = pathlib.Path(trainingDataDir)
        self.species = species
        self.datasetType = datasetType
        self.transforms = transforms
        self.includeWeights = includeWeights
        # construct all the file names  for all the species in the images, species dir will also have an
        # empty dir, containing empty channels, don't worry about this
        self.phaseDirs = []
        self.maskDirs = []
        self.weightDirs = []
        for dirName in self.species:
            self.phaseDirs.append(self.trainingDataDir.joinpath(dirName , 'phase_' + self.datasetType))
            self.maskDirs.append(self.trainingDataDir.joinpath(dirName, 'mask_' + self.datasetType))
            if self.includeWeights:
                self.weightDirs.append(self.trainingDataDir.joinpath(dirName, 'weights_' + self.datasetType))

        self.phaseFilenames = []
        self.maskFilenames = []
        self.weightFilenames = []

        for i, directory in enumerate(self.phaseDirs, 0):
            # grab filenames and add the corresponding mask to the file list
            filenames = [filename.name for filename in directory.glob('*.tif')]  
            for filename in filenames:
                # adding phaseFilenames
                self.phaseFilenames.append(directory.joinpath(filename))
                # adding maskFilenames
                self.maskFilenames.append(self.maskDirs[i].joinpath(filename))
                # adding weightFilenames
                if self.includeWeights:
                    self.weightFilenames.append(self.weightDirs[i].joinpath(filename))

    def __len__(self):
        return len(self.phaseFilenames)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.to_list()

        phaseImg = io.imread(self.phaseFilenames[idx])
        maskImg = io.imread(self.maskFilenames[idx])

        if self.includeWeights:
            weightImg = io.imread(self.weightFilenames[idx])
            weightFilename = self.weightFilenames[idx]
        else:
            weightImg = np.zeros(phaseImg.shape)
            weightFilename = None

        sample = {'phase': phaseImg, 'mask': maskImg, 'weights': weightImg,
                'phaseFilename': self.phaseFilenames[idx], 'maskFilename': self.maskFilenames[idx],
                'weightsFilename': weightFilename}
        
        if self.transforms != None:
            sample = self.transforms(sample)

        return sample

    # This will plot the data point that goes into the training net
    def plotDataPoint(self, idx):
        if torch.is_tensor(idx):
            idx = idx.to_list()

        if type(idx) == list:
            print("Plotter only works with integer indices, can only plot one item at a time :( ")
            return
        else:
            fig, (ax1, ax2, ax3) = plt.subplots(nrows=1, ncols=3)
            data = self.__getitem__(idx)
            ax1.imshow(data['phase'].numpy().squeeze(0), cmap='gray')
            #ax1.set_xlabel(str(data['phaseFilename']))
            ax1.set_xlabel('Phase')
            #ax1.set_title('Phase')

            ax2.imshow(data['mask'].numpy().squeeze(0), cmap='gray')
            #ax2.set_xlabel(str(data['maskFilename']))
            ax2.set_xlabel('Binary mask')
            #ax2.set_title('mask')

            ax3.imshow(data['weights'].numpy().squeeze(0))
            #ax3.set_xlabel(str(data['weightFilename']))
            ax3.set_xlabel('Weight map')
            #ax3.set_title('weights')
            plt.show()






import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

class basicUnet(nn.Module):
    '''
    Basic Unet stuff, simply copied the architecture with
    changes to sizes and layers

    
    '''

    def __init__(self, transposeConv = True):
        # initialize the superclass initializer from the library modules
        super(basicUnet, self).__init__()
        self.transposeConv = transposeConv
        # 1 because of the number of input channels is 1
        self.initial = double_conv(1, 64)
        self.down1 = double_conv(64, 128)
        self.down2 = double_conv(128, 256)
        self.down3 = double_conv(256, 512)
        self.down4 = double_conv(512, 512)

        self.up1 = up_conv_cat(1024, 256, self.transposeConv)
        self.up2 = up_conv_cat(512, 128, self.transposeConv)
        self.up3 = up_conv_cat(256, 64, self.transposeConv)
        self.up4 = up_conv_cat(128, 64, self.transposeConv)
        self.out = nn.Conv2d(64, 1, 1) # 2 because of no_classes

    def forward(self, x):
        # x will be the image batch tensor, that will be propagated across
        # toward the end and the 
        x1 = self.initial(x)
        x2 = self.down1(F.max_pool2d(x1, (2, 2)))
        x3 = self.down2(F.max_pool2d(x2, (2, 2)))
        x4 = self.down3(F.max_pool2d(x3, (2, 2)))
        x5 = self.down4(F.max_pool2d(x4, (2, 2)))

        # copied code
        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        x = self.out(x)


        return x
        #return torch.sigmoid(x)


class double_conv(nn.Module):
    '''
    Combining conv1- batch - relu - conv batch relu parts of UNet
    
    Features: No change in size of the image (maintained using padding)
    '''
    
    def __init__(self, input_channels, output_channels):
        ''' Just initializes the conv - batch - relu layers twice 
            with no reduction in image size (don't change padding , stride)
        '''
        super(double_conv, self).__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(input_channels, output_channels, 3, stride = 1, padding = 1),
            nn.BatchNorm2d(output_channels),
            nn.ReLU(inplace = True),
            nn.Conv2d(output_channels, output_channels, 3, stride = 1, padding =1),
            nn.ReLU(inplace = True)
        )

    def forward(self, x):
        x = self.conv(x)
        return x



class up_conv_cat(nn.Module):
    '''
    UpConv + concatenation of features during downscaling
    '''

    def __init__(self, input_channels, output_channels, transposeConv):
        super(up_conv_cat, self).__init__()

        if transposeConv:
            self.up = nn.ConvTranspose2d(input_channels//2, input_channels//2, 2, stride=2)
        else:
            self.up = nn.Upsample(scale_factor=2, mode = 'bilinear', align_corners=True)

        self.conv = double_conv(input_channels, output_channels)

    def forward(self, from_bottom, from_side):

        x1 = self.up(from_bottom)

        x2 = torch.cat([from_side, x1], dim = 1)

        x2 = self.conv(x2)
 
        return x2

class smallerUnet(nn.Module):
    '''
    Basic Unet stuff, simply copied the architecture with
    changes to sizes and layers, this one wont get 
    inputs from the side

    I am guessing they act a gradient highways preventing
    the net from learning features
    
    '''

    def __init__(self, transposeConv):
        # initialize the superclass initializer from the library modules
        super(smallerUnet, self).__init__()
        self.transposeConv = transposeConv
        # 1 because of the number of input channels is 1
        self.initial = double_conv(1, 8)
        self.down1 = double_conv(8, 16)
        self.down2 = double_conv(16, 32)
        self.down3 = double_conv(32, 64)
        self.down4 = double_conv(64, 64)

        self.up1 = up_conv(128, 32, self.transposeConv)
        self.up2 = up_conv(64, 16, self.transposeConv)
        self.up3 = up_conv(32, 8, self.transposeConv)
        self.up4 = up_conv(16, 8, self.transposeConv)
        self.out = nn.Conv2d(8, 1, 1) # 2 because of no_classes

    def forward(self, x):
        # x will be the image batch tensor, that will be propagated across
        # toward the end and the 
        x1 = self.initial(x)
        x2 = self.down1(F.max_pool2d(x1, (2, 2)))
        x3 = self.down2(F.max_pool2d(x2, (2, 2)))
        x4 = self.down3(F.max_pool2d(x3, (2, 2)))
        x5 = self.down4(F.max_pool2d(x4, (2, 2)))

        # copied code
        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        x = self.out(x)


        return x # sigmoid is done in the loss function

class up_conv(nn.Module):
    '''
    UpConv and no concatenation from the side, pure down convs and up-convs
    '''

    def __init__(self, input_channels, output_channels, transposeConv):
        super(up_conv, self).__init__()

        if transposeConv:
            self.up = nn.ConvTranspose2d(input_channels//2, input_channels//2, 2, stride=2)
        else:
            self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)

        self.conv = double_conv(input_channels, output_channels)

    def forward(self, from_bottom, from_side):

        x1 = self.up(from_bottom)

        x2 = torch.cat([from_side, x1], dim = 1)

        x2 = self.conv(x2)
 
        return x2



from scipy.ndimage.morphology import distance_transform_edt
import torch
import torch.nn.functional as F
import numpy as np
import glob
import os
from skimage.io import imread
from skimage.measure import label, regionprops
from skimage.filters import threshold_otsu
from scipy.ndimage import gaussian_filter
from narsil.utils.growth import exp_growth_fit
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle

class singleChannelFishData(object):
	"""
	Class that bundles the FISH data from all the imaging channels and
	generate bounding boxes around the signal and also does background
	subtraction

    Arguments:
    channelNames = list of channel names eg. ['488', 'cy3', 'cy5', 'txr']
    transforms = type of transform you want applied on the fluor images
    fluorChThreshold = list of values for thresholding individual flourChannels
    stdThreshold = no of deviations above which the the flour signal is thresholded in
                   mean-std-normalization
    
    """
    def __init__(self, fishDirName, channelNames, channelSize = (60, 750), minboxHeight = 90, transforms = None, 
                    fluorChThreshold = None, stdThreshold = 1.0, fileFormat = '.tiff',
                    backgroundFishData = None):
        self.fishDirName = fishDirName
        self.channelNames = channelNames
        self.fileFormat = fileFormat
        self.stdThreshold = stdThreshold
        self.fishImages = {}
        self.fishBoxes = {}
        self.fluorChThreshold = fluorChThreshold
        self.channelSize = channelSize
        self.transforms = transforms
        self.minboxHeight = minboxHeight

        # pass in a data object of class singleChannelFishData of the background
        # channel for subtraction
        self.backgroundFishData = backgroundFishData
        for channel in channelNames:
            image = imread(self.fishDirName + channel + self.fileFormat, as_gray=True)
            #image = (image - np.min(image))/(np.max(image) - np.min(image)) * 65535
            if transforms == 'mean-std-normalization':
                self.fishImages[channel]  = (image - np.mean(image))/np.std(image) > stdThreshold

            # basic idea to be able to draw a box around the region where the flour channels are
            elif transforms == 'box':
                #image = gaussian_filter(image, sigma = 4)
                #background subtraction
                image = image - self.backgroundFishData[channel]
                image = gaussian_filter(image, sigma = 2)
                #print("Image found")
                if np.sum(image) == 0:
                    self.fishImages[channel] = self.backgroundFishData[channel]
                    self.fishBoxes[channel] = []
                else:
                    image[:self.channelSize[0], :] = 0
                    image[self.channelSize[1]:, :] = 0
                    #image = (image - np.min(image))/(np.max(image) - np.min(image)) * 65535
                    #image = image - np.mean(image)
                    image_bool = image > max(threshold_otsu(image), self.fluorChThreshold[channel])
                    boxes = []
                    y1 = 5
                    y2 = image.shape[1] - 5
                    width = y2 - y1 + 1
                    xlims_bool = np.sum(image_bool, axis=1) > 0
                    xlims = np.where(np.diff(xlims_bool) == 1)[0]
                    #print(f"{channel}: --> xlims: {xlims}")
                    if len(xlims)%2 == 1 and len(xlims) != 0:
                        xlims = xlims[:-1]
                    for i in range(0, len(xlims), 2):
                        xy = (y1, xlims[i])
                        height = xlims[i+1] - xlims[i] + 1
                        if height >= self.minboxHeight:
                            boxes.append((xy, width, height))
                    #print(f"{channel} : {boxes}")
                    if len(boxes) != 0:
                        self.fishBoxes[channel] = boxes
                    else:
                        self.fishBoxes[channel] = []
                    self.fishImages[channel] = image

            else:
                self.fishImages[channel] = image

        # kinda apply transforms to clean up individual flourescent channels and 
        # make them binary masks to mapy correctly

    def __len__(self):
        return len(self.channelNames)

    def __getitem__(self, channel):
        if channel not in self.channelNames:
            return None
        else:
            return self.fishImages[channel]

    def plot(self, channel):
        fig, ax = plt.subplots(nrows = 1, ncols = 1)
        ax.imshow(self.fishImages[channel], cmap='gray')
        if channel in self.fishBoxes.keys():
            for box in self.fishBoxes[channel]:
                ax.add_patch(Rectangle(*box, linewidth = 1, edgecolor='r', facecolor ='none'))
        ax.set_title(f'{channel} Image ')
        plt.show()

    # This will count the number of channels in which the boxes are drawn 
    # around the fluor pixels in the images.
    def getNumberFluorChannels(self):
        numberChannels = 0
        for channel in self.channelNames:
            if len(self.fishBoxes[channel]) > 0:
                numberChannels += 1
        return numberChannels

    def getFluorChannels(self):
        fluorChannels = []
        for channel in self.channelNames:
            if len(self.fishBoxes[channel]) > 0:
                fluorChannels.append(channel)
        return fluorChannels

    def plotAllChannels(self):
        fig, ax = plt.subplots(nrows = 1, ncols = len(self.channelNames))
        for i, channel in enumerate(self.channelNames, 0):
            ax[i].imshow(self.fishImages[channel], cmap='gray')
            if self.transforms == 'box' and len(self.fishBoxes[channel]) != 0:
                for box in self.fishBoxes[channel]:
                    ax[i].add_patch(Rectangle(*box, linewidth = 1, edgecolor = 'r', facecolor = 'none'))
            ax[i].set(title=channel)
        plt.show(block = False)
        fig.canvas.set_window_title(f"{self.fishDirName}")


class singlePositionFishData(object):
    """
    Class for holding fluorescent image of one position and also helps in 
    a background subtraction and extraction of individual channels from 
    channel locations

    """
    
    def __init__(self, fishDir, channelNames, locations, 
            channelWidth=40, imgName='img_000000000', fileformat='.tiff'):
        self.fishDir = fishDir
        self.channelNames = channelNames
        self.locations = locations #locations of the mother machine channels
        self.channelWidth = channelWidth
        self.fileformat = fileformat
        self.fishImages = {}

        # set images
        for channel in self.channelNames:
            image = imread(fishDir + '/' + str(channel) + '/' + imgName + self.fileformat, as_gray=True)
            self.fishImages[channel] = image

    def __len__(self):
        return len(self.channelNames)
    
    def __getitem__(self, channel):
        if channel not in self.channelNames:
            return None
        else:
            return self.fishImages[channel]
    
    def generateBboxes(self):
        pass
    
    def subtractBackground(self):
        pass

    def plotChannel(self, channel, withBboxes=True):
        pass



# Network architecture of the nets used for tracking
import torch
import torch.nn as nn
import torch.nn.functional as F


class siameseNet(nn.Module):

    def __init__(self, outputFeatureSize):
        super(siameseNet, self).__init__()
        # Take 11 input properties and then make 10 vector
        self.outputFeatureSize = outputFeatureSize
        self.propertiesHead = nn.Sequential(
            nn.Linear(11, 200),
            nn.ReLU(inplace=True),
            nn.Linear(200, 100),
            nn.ReLU(inplace=True),
            nn.Linear(100, self.outputFeatureSize)
        )
        # Do convolutions and then fully connected layer and 
        # make a 10 vector
        # image shape is set in the dataLoader, not here 
        # imgStdsize = (96,32) 
        self.imageHead = nn.Sequential(
                            nn.Conv2d(1, 4, kernel_size=(3,3), stride = 1, padding = 1),
                            nn.ReLU(inplace=True),
                            nn.BatchNorm2d(4),
                            nn.Dropout2d(p=0.2),

                            nn.Conv2d(4, 8, kernel_size=(3,3), stride=1, padding=1),
                            nn.ReLU(inplace=True),
                            nn.BatchNorm2d(8),
                            nn.Dropout2d(p=0.2),

                            nn.Conv2d(8, 8, kernel_size=(3,3), stride=1, padding=1),
                            nn.ReLU(inplace=True),
                            nn.BatchNorm2d(8),
                            nn.Dropout2d(p=0.2),

                            )
        # substitute the * with image dimension
        self.fc1imageHead = nn.Sequential(
                                nn.Linear(8 * 96 * 32, 500),
                                nn.ReLU(inplace=True),
                                nn.Linear(500, 500),
                                nn.ReLU(inplace=True),
                                nn.Linear(500, self.outputFeatureSize)
                            )

        
    # Applies the two network heads and concatenates the output
    def forward_oneSet(self, prop, image):
        propertiesHeadOutput = self.propertiesHead(prop)
        imghead = self.imageHead(image)
        fc1input = imghead.view(imghead.size()[0], -1)
        imageHeadOutput = self.fc1imageHead(fc1input)
        return torch.cat((propertiesHeadOutput, imageHeadOutput), 1) # Careful here, this is tested to work
    
    # This will take two inputs and spit out one output
    # which is something like a distance metric, in the 
    # space of similartiy or not.
    def forward(self, prop1, image1, prop2, image2):
        output1 = self.forward_oneSet(prop1, image1)
        output2 = self.forward_oneSet(prop2, image2)
        return output1, output2


# Functions that can bundle channels from different positions 
import numpy as np
import glob
import os
import time
import pickle
import torch
import torch.nn as nn
import torch.nn.functional as F
import multiprocessing as mp
import shutil
from torch.utils.data import Dataset, DataLoader
from narsil.fish.datasets import singleChannelFishData
from narsil.tracking.siamese.network import siameseNet
from narsil.tracking.siamese.datasets import singleChannelTrackingSiamese
from functools import partial
from multiprocessing import Pool

def evalifyNet(modelPath, device):
	device = torch.device(device if torch.cuda.is_available() else "cpu")
	
	savedModel = torch.load(modelPath)
	net = siameseNet(outputFeatureSize=savedModel['modelParameters']['outputFeatureSize'])
	net.share_memory()
	net.load_state_dict(savedModel['model_state_dict'])
	net.to(device)
	net.eval()
	for p in net.parameters():
		p.requires_grad_(False)
	
	return net

# Main function that processes tracks from one channel and calculated growth rates and species ID
def processOneChannel(channelDirName, net, bgFishData, trackingParameters):

	device = torch.load(trackingParameters['device'] if torch.cuda.is_available() else "cpu")
	batch_size = trackingParameters['batch_size']
	fluorChThreshold = trackingParameters['fluorChThreshold']
	channelSize = trackingParameters['channelSize']

	oneChannel = singleChannelTrackingSiamese(channelDirName, trackingParameters=trackingParameters,
				backgroundFishData=bgFishData, fluorChThreshold=fluorChThreshold, channelSize=channelSize)
	#print(len(oneChannel))
	# num_workers = 0 as the number of copies of the net are running at same time at process level
	# don't change it 
	stackdataloader = DataLoader(oneChannel, batch_size=batch_size, shuffle=False, num_workers=0)

	# results to keep track of scores
	prevFileNames = []
	currFileNames = []
	prevBlobNumbers = []
	currBlobNumbers = []
	scores = []

	# Run through the net and accumulte the results
	with torch.no_grad():
		for i, data in enumerate(stackdataloader, 0):
			blob1_props, blob1_image, blob2_props, blob2_image = data[0].to(device), data[1].to(device), data[2].to(device), data[3].to(device)
			out1, out2 = net(blob1_props, blob1_image, blob2_props, blob2_image)
			net_scores = F.pairwise_distance(out1, out2, keepdim=True)
			prevFileNames.extend(data[4])
			prevBlobNumbers.extend(data[5].numpy())
			currFileNames.extend(data[6])
			currBlobNumbers.extend(data[7].numpy())
			output_scores = net_scores.to("cpu").detach().numpy()
			scores.extend(output_scores)
	# return prevFileNames, currFileNames, prevBlobNumbers, currBlobNumbers, scores	

	# Data goes into tracking above only if there are sufficent no of fluorescent channels lighting up 
	# you can always call oneChannel.createDataForNet() to let data for tracking be created
	if oneChannel.doneTracking:
		oneChannel.constructLinks(prevFileNames, currFileNames, prevBlobNumbers, currBlobNumbers, scores)
		oneChannel.convertLinksToTracks()
		oneChannel.labelTracksWithFluorChannels(printFluorLabels=False)
		oneChannel.setSpeciesForAllTracks()

		localGrowthRates = oneChannel.getGrowthOfSpeciesLabeledTracks()
		if trackingParameters['plot'] == True:
			#oneChannel.plotAllTracksWithFISH()
			oneChannel.plotTracksUsedForGrowth()

		if trackingParameters['writeGrowthRates'] == True:
			writeDirectory = channelDirName.split('blobs')[0] + str('growthRates/')
			channelNumber = channelDirName.split('blobs')[1].strip('/') # still a string
			if not os.path.exists(writeDirectory):
				os.makedirs(writeDirectory)
			
			writeFilename = writeDirectory + channelNumber + ".pickle"

			with open(writeFilename, "wb") as handle:
				pickle.dump(localGrowthRates, handle, protocol=pickle.HIGHEST_PROTOCOL)
	
	del oneChannel

def processOnePosition(positionDirName, net, trackingParameters):
	
	positionNumber = int(positionDirName.split('blobs')[0].split('/')[-2][3:])

	# we use background channel to be able to subtract background from the fish data
	if trackingParameters['backgroundChannelNumber'] != None and (positionNumber not in trackingParameters['flipPositions']):
		# set background fish directory
		backgroundFishDir = positionDirName.split('blobs')[0] + 'fishChannels/' + str(trackingParameters['backgroundChannelNumber']) + '/'
		bgFishData = singleChannelFishData(backgroundFishDir, channelNames=trackingParameters['channelNames'], transforms=None)
	elif trackingParameters['backgroundChannelNumber'] != None and (positionNumber in trackingParameters['flipPositions']):
		# set background fish directory
		backgroundFishDir = positionDirName.split('blobs')[0] + 'fishChannels/' + str(trackingParameters['backgroundChannelNumber'] + 1) + '/'
		bgFishData = singleChannelFishData(backgroundFishDir, channelNames=trackingParameters['channelNames'], transforms=None)
	else:
		bgFishData = None

	for i in range(trackingParameters['numChannels']):
		channeldirName = positionDirName + str(i) + '/'
		processOneChannel(channeldirName, net, bgFishData, trackingParameters=trackingParameters)

	print(f"{positionDirName.split('blobs')[0]} -- Done")
	

def processAllPositions(analysisDir, positions, trackingParameters, 
						skipPositions, numProcess=6):
	start = time.time()
	listPositionDirs = []
	for position in positions:
		if position not in skipPositions:
			listPositionDirs.append(analysisDir + 'Pos' + str(position) + '/blobs/')

	print(listPositionDirs)
	trackingnet = evalifyNet(trackingParameters['modelPath'], trackingParameters['device'])

	try:
		mp.set_start_method('spawn')
	except RuntimeError:
		pass

	pool = mp.Pool(processes=numProcess)
	pool.map(partial(processOnePosition, net=trackingnet, trackingParameters=trackingParameters), listPositionDirs)
	pool.close()
	pool.join()

	duration = time.time() - start

	print(f"Duration of tracking {len(positions)} positions is {duration}s")

def deleteTrackingData(analysisDir, positions, dirNameToDelete):

	dirsToDelete = []
	for position in positions:
		directory = analysisDir + 'Pos' + str(position) + '/' + dirNameToDelete + '/'
	
	for directory in dirsToDelete:
		try:
			shutil.rmtree(directory)
		except:
			print(f"Eror when deleting {directory} -- probably doesn't exist")



# File containing datasets used for training the siamese net
import glob
import torch
import numpy as np
import random
import matplotlib.pyplot as plt
from torch.utils.data import Dataset, DataLoader
from skimage.io import imread, imsave
from skimage.measure import label, regionprops
from skimage.transform import resize
from collections import OrderedDict


class channelDataset(object):

    def __init__(self, directory, labelled= False, frameskip=1, fileformat='.tiff'):

        self.directory = directory
        self.fileformat = fileformat
        self.labelled = labelled
        self.n_images = len(glob.glob(self.directory + "*" + self.fileformat))
        self.indices = [int(filename.split('.')[0].split('/')[-1]) for filename in 
                        glob.glob(self.directory + "*" + self.fileformat)]
        self.indices.sort()
        self.indices = self.indices[::frameskip]

    def __len__(self):
        return len(self.indices)
    
    def __getitem__(self, idx):

        channel_img_name =  self.directory + str(self.indices[idx]) + self.fileformat
        channel_img = imread(channel_img_name)
        if self.labelled:
            channel_img = label(channel_img, connectivity=2) 
        return channel_img


class siameseDatasetWrapper(Dataset):

    def __init__(self, data, transforms=None):
        self.data = data
        self.transforms = transforms

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        sample = self.data[idx]

        if self.transforms != None:
            sample = self.transforms(sample)

        return sample

class siameseDataset(object):

    def __init__(self, dirNames, inlcudeDaughters=True, imgStdSize=(96, 32),
                 validation=True,
                 fileformat='.tiff',linksformat='.npy', frameskip=1):
        self.dirNames = dirNames
        self.imgStdSize = imgStdSize
        self.fileformat = fileformat
        self.linksformat = linksformat
        self.frameskip = frameskip
        self.includeDaughters = inlcudeDaughters
        self.validation = validation

        self.data = []

        self.dataInclDaughterLinks = []

        self.trainingData = []
        self.validationData = []

        for directory in self.dirNames:
            tracksSet = setOfTracksTrain(directory, fileformat=self.fileformat,
                                    linksformat=self.linksformat, frameskip=self.frameskip)
            linksFileName = directory + 'justlinks' + self.linksformat
            linksList = np.load(linksFileName).tolist()
            #numberOfFiles = len(tracskSet.indices)
            fileIndices = tracksSet.indices
            #for i in range(len(tracksSet)):
            #    tracksSet.printTrack(i)
            #    print('----------------------')

            # Iterate through the frames and construct data object
            # Better to load them from memory on the fly
            for index in range(len(fileIndices) - 1):
                index_t = fileIndices[index]
                fileName_t = directory + str(index_t) + self.fileformat
                index_t1 = fileIndices[index + 1]
                fileName_t1 = directory + str(index_t1) + self.fileformat
                # Loop through the blobs and add them to data
                imgBundle1 = imgPropBundle(fileName_t, self.imgStdSize)
                imgBundle2 = imgPropBundle(fileName_t1, self.imgStdSize)
                nBlobs1 = len(imgBundle1)
                nBlobs2 = len(imgBundle2)
                # Check the tracksSet elements for index_t: blob_t and index_t1: blob_t1
                # Loop through each set of tracks and check if both exists
                for blob1 in range(nBlobs1):
                    for blob2 in range(nBlobs2):
                        # loop through each of the tracks and check
                        blobs_connected_incl_daughters = False
                        if [index_t, blob1, blob2] in linksList:
                            blobs_connected_incl_daughters = True
                        
                        blobs_connected = False
                        for track in tracksSet:
                            # if both blobs exist in one track, the set the label to 1 as they are linked
                            # if not then set the label to 0, as they are not linked
                            if((index_t in track and blob1 == track[index_t]) and (index_t1 in track and blob2 == track[index_t1])):
                                blobs_connected = True
                                break
                        if (blobs_connected):
                            item = [imgBundle1[blob1], imgBundle2[blob2], True]
                            #print(item)
                            self.data.append(item)
                        elif (not blobs_connected):
                            item = [imgBundle1[blob1], imgBundle2[blob2], False]
                            #print(item)
                            self.data.append(item)
                        if (blobs_connected_incl_daughters):
                            item = [imgBundle1[blob1], imgBundle2[blob2], True]
                            self.dataInclDaughterLinks.append(item)
                        elif (not blobs_connected_incl_daughters):
                            item = [imgBundle1[blob1], imgBundle2[blob2], False]
                            self.dataInclDaughterLinks.append(item)
            
        self.getLinkedData()
        self.getNotLinkedData()
        numberLinkedData = len(self.linkedData)
        self.sampledNonLinkedData = random.sample(self.notLinkedData, numberLinkedData)
        self.balancedData = self.linkedData + self.sampledNonLinkedData
    
    def __len__(self):
        return len(self.balancedData)

    def splitData(self, p=0.8):
        trainData = []
        validationData = []
        if self.validation == True:
            for data in self.balancedData:
                if random.random() >= 1 - p:
                    trainData.append(data)
                else:
                    validationData.append(data)
        else:
            trainData = self.balancedData
        
        return trainData, validationData


    def getLinkedData(self):
        if self.includeDaughters:
            data = self.dataInclDaughterLinks
        else:
            data = self.data
        self.linkedData = []
        for item in data:
            if item[2] == True:
                self.linkedData.append(item)
    
    def getNotLinkedData(self):
        if self.includeDaughters:
            data = self.dataInclDaughterLinks
        else:
            data = self.data
        self.notLinkedData = []
        for item in data:
            if item[2] == False:
                self.notLinkedData.append(item)



class setOfTracksTrain(Dataset):
    """
    A class holding a set of tracks for a channel

    TODO: DONT use frameskip option, it is not fully implemented in all functions, in the current structure 
    """
    def __init__(self, directory, fileformat = '.tiff', linksformat = '.npy',frameskip = 1):
        self.directory = directory
        self.fileformat = fileformat 
        self.frameskip = frameskip
        self.indices = [int(filename.split('/')[-1].split('.')[0]) for filename in glob.glob(self.directory + "*" + self.fileformat)]
        self.indices.sort()
        self.indices = self.indices[::self.frameskip]
        # Data is coming from the labeled dataset
        self.data = channelDataset(self.directory, labelled= True, frameskip=self.frameskip, fileformat=self.fileformat)
        self.unlabeleddata = channelDataset(self.directory, labelled=False, frameskip=self.frameskip, fileformat=self.fileformat)
        self.linksformat = linksformat
        self.linksFileName = self.directory + 'justlinks' + self.linksformat

        # Load the tracks from numpy dictionary here
        self.loadLinksFromNumpy()
        self.props = []
        for i in range(len(self.data)):
            regionpropsOneImage = regionprops(self.data[i])
            self.props.append(regionpropsOneImage)

        # Tracks is a list, could be a set. will refactor later
        # Each element is a track, which is a dictionary containing keys(Frame numbers)
        # and values(blob numbers), and any other details can be added to the values for
        # indexing into later, like quality labels, time stamps, etc
        self.constructTracks()

    def __len__(self):
        # Return the total number of tracks after loading 
        return len(self.tracks)

    def __str__(self):
        return 'Number of tracks are' + str(self.__len__())

    def __getitem__(self, idx):
        # function returns a set of images
        if torch.is_tensor(idx):
            idx = idx.tolist()
        # sort the dictionary items if needed or just in case, 
        sorted_track = OrderedDict(sorted(self.tracks[idx].items()))
        return sorted_track

    def printTrack(self, idx):
        # Funciton that prints the details of one track to terminal to read
        track = self.__getitem__(idx)
        # sort the the dictionary by time and then print the track
        for key, value in track.items():
            print("Frame No: ", key, " Blob No: ", value, "bbox: ", self.props[key][value]['bbox'])
            
    
    def plotTrack(self, idx):
        # Function to plot and track blobs in matplotlib
        track = self.__getitem__(idx)
        track_length = len(track)
        # convert for iteration
        track_list = [(frame_t, blob_t) for frame_t, blob_t in track.items()]
        fig, axs = plt.subplots(1, track_length)
        i = 0
        for frame_t, blob_t in track_list:
            axs[i].imshow(self.props[frame_t][blob_t]['image'])
            #axs[i].title(str(frame_t))
            i += 1
        

    def loadLinksFromNumpy(self):
        # Here iterate over the matrix and figure out the appropriate blobs 
        # You would have to construct the track iteratively as there is no easy way 
        # to access a particular track using a particular blob without looping.
        self.linksArray = np.load(self.linksFileName)

    def plotAllTracksFromNumpy(self, spacing = 0):
        # Function plots all the track links read from numpy training matrix
        n_images = len(self.unlabeleddata)
        print("Number of images: ", n_images)
        if n_images > 0: 
            height, width = self.unlabeleddata[0].shape
            print("Height: ", height, " Width: ", width)
        full_img = np.zeros(shape=(height, n_images * (width + spacing)))
        indices = self.unlabeleddata.indices
        print("Full image shape: ", full_img.shape)

        # set the images
        for i in range(n_images):
            sub_image = self.unlabeleddata[i]
            spacing_image = np.zeros(shape=(height, spacing))
            sub_image = np.append(sub_image, spacing_image, axis = 1)
            full_img[:, i*(width + spacing) : (i+1) * (width+spacing)] = sub_image

        plt.figure()
        plt.imshow(full_img)

        # From the linksArray, grab the stuff from the frame numbers in indices
        # 
        for row_index in range(self.linksArray.shape[0]):
            frame_t, blob_t, blob_t1 = self.linksArray[row_index]
            centroid_t_x, centroid_t_y = self.props[frame_t][blob_t]['centroid']
            centroid_t1_x, centroid_t1_y = self.props[frame_t + 1][blob_t1]['centroid']
            plt.plot([centroid_t_y + frame_t*(width + spacing), centroid_t1_y + (frame_t + 1)*(width + spacing)], [centroid_t_x, centroid_t1_x], 'r')

        plt.title('Links from Numpy traininig matrix')
        plt.show(block = False)
    
    def constructTracks(self):
        # This is a function that constructs tracks from the links Matrix.
        # Iterate till you go no where or have two blobs connected to one blob
        # 
        self.tracks = [] # a list of dictionaries, each track is a dictionary
        # keys are frame numbers and values are indices of the blobs
        # create local copies tot not change the classes internal ones
        linksArray = self.linksArray
        
        while linksArray.size != 0:
        #for row in range(linksArray.shape[0]):
            # start with one element and end when the track ends
            # or when there are two blobs connecting the same blob
            first_row = linksArray[0] # start linking from here
            one_track = {first_row[0] : first_row[1]} # add the first blob
            # Now loop till there is nothing matching
            # How many blobs is this blob connected to : this blob means the first row
            while True:
                current_blob_index = np.where((linksArray[:, :2] == [first_row[0], first_row[1]]).all(axis=1))[0]
                #print("Current_blob_index: ", current_blob_index)
                connections_from_current_blob = np.where((linksArray[:, :2] == [first_row[0], first_row[1]]).all(axis=1))[0]
                if(len(connections_from_current_blob) == 2):
                    # Time to split the track and not add the track, find out which are daughters and keep track
                    #print("Blob has daughter and split: ", first_row, connections_from_current_blob)
                    linksArray = np.delete(linksArray, (current_blob_index), axis = 0)
                    break
                next_blob_index = np.where((linksArray[:,:2] == [first_row[0] + 1, first_row[2]]).all(axis=1))[0] # Grab the index of the next blob from the array
                if next_blob_index.size == 0:
                    #print("Track ended break")
                    linksArray = np.delete(linksArray, (current_blob_index), axis = 0)
                    break
                else:
                    next_row = linksArray[next_blob_index[0]]
                    one_track[next_row[0]] = next_row[1]
                    linksArray = np.delete(linksArray, (current_blob_index), axis = 0)

                first_row = next_row
            #print(one_track)
            self.tracks.append(one_track)



class imgPropBundle(object):

    def __init__(self, imgFilename, imgStdSize=(96, 32)):
        self.imgFilename = imgFilename
        self.imgStdSize = imgStdSize
        self.img = imread(self.imgFilename)
        # props
        self.props = regionprops(label(self.img))
    
    def __len__(self):
        return len(self.props)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        # accumulate properties you need for the image encoder net
        local_blob_prop = self.props[idx]
        # Use a torch tensor may be
        # bbox, centroid, area, length, ellipsicity
        bbox = list(local_blob_prop['bbox'])
        area = local_blob_prop['area']
        centroid = list(local_blob_prop['centroid'])
        eccentricity = local_blob_prop['eccentricity']
        major_axis_length = local_blob_prop['major_axis_length']
        minor_axis_length = local_blob_prop['minor_axis_length']
        perimeter = local_blob_prop['perimeter']
        image = local_blob_prop['image']

        # Do some operations on the image to center the image in the field,
        # to standardize the inputs to the network
        height, width = image.shape
        #print("Image shape before:", image.shape)
        # set blank background image

        # Generally wouldn't happen, just in-case safety measure
        # Should be redone in a different way than resizing
        height_outside_limit = height > self.imgStdSize[0]
        width_outside_limit = width > self.imgStdSize[1]

        if height_outside_limit:
            image = resize(image, (self.imgStdSize[0], width))
        if width_outside_limit:
            image = resize(image, (height, self.imgStdSize[1]))

        background = np.zeros(self.imgStdSize)

        # recalculate height after rescaling in case
        height, width = image.shape
        if height%2 == 1:
            xlim_1 = self.imgStdSize[0]//2 - height//2
            xlim_2 = self.imgStdSize[0]//2 + height//2 + 1
        elif height%2 == 0:
            xlim_1 = self.imgStdSize[0]//2 - height//2
            xlim_2 = self.imgStdSize[0]//2 + height//2

        if width%2 == 1:
            ylim_1 = self.imgStdSize[1]//2 - width//2
            ylim_2 = self.imgStdSize[1]//2 + width//2 + 1
        elif width%2 == 0:
            ylim_1 = self.imgStdSize[1]//2 - width//2
            ylim_2 = self.imgStdSize[1]//2 + width//2

        #print(xlim_1, xlim_2, ylim_1, ylim_2)
        #print("Image shape:", image.shape)
        
        background[xlim_1 : xlim_2, ylim_1: ylim_2] = image
        

        # Construct properties list
        properties = []
        properties.extend(bbox)
        properties.append(area)
        properties.extend(centroid)
        properties.append(eccentricity)
        properties.append(major_axis_length)
        properties.append(minor_axis_length)
        properties.append(perimeter)

        return { 'props' : properties, 'image' : background}

    def plotImage(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        plt.figure()
        plt.imshow(self.__getitem__(idx)['image'])
        plt.show(block = False)

class oneSet(Dataset):
    """ This will  load a set of images for tracking training data
    One folder containes all the images of a channels.
    This class holds all the information together and helps in 
    loading on the GUI
    """

    def __init__(self, directory, transforms = None, fileformat = '.tiff'):

        self.directory = directory
        self.fileformat = fileformat
        self.transforms = transforms
        self.indices = [int(filename.split('/')[-1].split('.')[0]) for filename in glob.glob(self.directory + "*" + self.fileformat)]
        self.indices.sort()
        self.n_timepoints = len(glob.glob(self.directory + "*" + self.fileformat))

    def __len__(self):
        return self.n_timepoints

    def __str__(self):
        return 'directory: {self.directory} set'.format(self=self)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        filename = self.directory + str(self.indices[idx]) + self.fileformat
        #print("Filename: ", filename)

        image = imread(filename, as_gray=True).astype('float32')
        label_image = label(image)
        regions = regionprops(label_image)
        bbox_list = []
        for prop in regions:
            bbox_list.append(prop.bbox)
        

        if self.transforms:
            image = self.transforms(image)

        return {'filename': filename, 'bbox': bbox_list, 'props': regions}



# File containing various loss functions
import torch
import torch.nn as nn
import torch.nn.functional as F


class WeightedUnetLoss(nn.Module):
	"""
	Custom loss function for Unet BCE + DICE + weighting
	"""

	def __init__(self):
		super(WeightedUnetLoss, self).__init__()

	def forward(self, output, target, weights):

		output = torch.sigmoid(output)

		batch_size = target.shape[0]


		target_weighted = target - weights
	
		output_reshaped = output.view(batch_size, -1)
		target_weighted_reshaped = target_weighted.view(batch_size, -1)



		bce_loss = F.binary_cross_entropy(output_reshaped, target_weighted_reshaped)

		target_reshaped = target.view(batch_size, -1)

		intersection = (output_reshaped * target_reshaped)
		dice_per_image = 2. * (intersection.sum(1)) / (output_reshaped.sum(1) + target_reshaped.sum(1))
		dice_batch_loss = 1 - dice_per_image.sum() / batch_size
		print(dice_batch_loss.item(), bce_loss.item())
		return 0.5 * bce_loss  + 2.0 *dice_batch_loss
	#return bce_loss


class WeightedUnetLossExact(nn.Module):
	"""
	Custom loss function that implements the exact formulation of weighted BCE in 
	from the U-net paper https://arxiv.org/pdf/1505.04597.pdf
	"""

	def __init__(self):
		super(WeightedUnetLossExact, self).__init__()

	def forward(self, output, target, weights):
		
		output = torch.sigmoid(output)

		batch_size = target.shape[0]

		# calculate intersection over union
		target_reshaped = target.view(batch_size, -1)
		output_reshaped = output.view(batch_size, -1)
		intersection = (output_reshaped * target_reshaped)
		dice_per_image = 2. * (intersection.sum(1)) / (output_reshaped.sum(1) + target_reshaped.sum(1))
		dice_batch_loss = 1 - dice_per_image.sum() / batch_size

		# weighted cross entropy function

		weights_reshaped = weights.view(batch_size, -1)  + 1.0
		bce_loss = weights_reshaped * F.binary_cross_entropy(output_reshaped, target_reshaped, reduction='none')

		weighted_entropy_loss = bce_loss.mean()		
		print(dice_batch_loss.item(), weighted_entropy_loss.item())
		return dice_batch_loss  + 0.5 * weighted_entropy_loss



class UnetLoss(nn.Module):
	"""
	Custom loss function, for Unet, BCE + DICE
	"""
	def __init__(self):
		super(UnetLoss, self).__init__()
		self.bce_loss = nn.BCELoss()

	def forward(self, output, target):

		output = torch.sigmoid(output)
		output_flat = output.view(-1)
		target_flat = target.view(-1)

		bce_loss = self.bce_loss(output_flat, target_flat)

		batch_size = target.shape[0]

		output_dice = output.view(batch_size, -1)
		target_dice = target.view(batch_size, -1)

		intersection = (output_dice * target_dice)
		dice_per_image = 2. * (intersection.sum(1)) / (output_dice.sum(1) + target_dice.sum(1))

		dice_batch_loss = 1 - dice_per_image.sum() / batch_size
		print(dice_batch_loss.item(), bce_loss.item())
		return 0.5 * bce_loss + dice_batch_loss
		#return bce_loss

class BCELoss(nn.Module):

	def __init__(self):
		super(BCELoss, self).__init__()
		self.bce_loss = nn.BCELoss()

	def forward(self, output, target):

		output = torch.sigmoid(output)

		output_flat = output.view(-1)
		target_flat = target.view(-1)

		bce_loss = self.bce_loss(output_flat, target_flat)

		return bce_loss


class ContrastiveLoss(nn.Module):
	"""
	Contrastive Loss function
	"""
	def __init__(self, margin=10.0):
		super(ContrastiveLoss, self).__init__()
		self.margin = margin

	def forward(self, output1, output2, label):
		euclidian_distance = F.pairwise_distance(output1, output2, keepdim=True)
		contrastive_loss = (1 - label) * torch.pow(euclidian_distance, 2) + label * torch.pow(torch.clamp(self.margin - euclidian_distance, min = 0.0), 2)
		contrastive_loss = torch.mean(contrastive_loss)
		return contrastive_loss


# Datasets loaders for tracking package,including training and testing and running the tracker
import glob
import torch
import numpy as np
from collections import OrderedDict
from skimage.measure import regionprops, label
from skimage.io import imread
from skimage.transform import resize
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from scipy.optimize import curve_fit
from narsil.utils.growth import exp_growth_fit
from narsil.utils.transforms import singleBlobTensorize
from narsil.fish.datasets import singleChannelFishData
from torch.utils.data import Dataset, DataLoader

np.seterr(over='ignore')



class oneCellLineage(object):
    """
    This class keeps track of one track, starting from birth to division
    and also attributes that help us locate its daughter tracks 
    """

    def __init__(self, dirName,exptTime = 61, fileformat = '.tiff'):
        self.dirName = dirName
        self.fileformat = fileformat
        # Tracks dictionary is something like (str) 'fileName': (int) blob number
        self.trackDictionary = OrderedDict()
        # is also a dictionary with something like (str) 'fileName': (int) blob number
        self.daughtersDictionary = OrderedDict()
        # is also a dicitonary with something like (str) 'fileName': (int) blob number

        self.fileBlobsDict = OrderedDict()
        self.parent = {}

        self.species = None
        self.fluorChannels = None
        # properties to grab areas or lengths
        self.props = {}
        self.areas = {}
        self.lengths = {}

        self.trackLength = 0
        self.exptTime = exptTime

        self.areas_np = -1 * np.ones((exptTime,))
        self.growth_np = -1 * np.ones((exptTime,))
        self.areas_rolling_np = -1 * np.ones((exptTime,))
        self.growth_rolling_np = -1 * np.ones((exptTime,))


        # These get set as internals and need to be reset if you by chance delete an
        # element form the list of tracks in one single channel.
        # These variables are set to link with daughters or parents and useful for 
        # iteration elsewhere
        self._indexToDaughters = []
        

    def __str__(self):
        # Here you print a representation of the lineage, tracks linked and
        # also daughters
        printStr = ""
        printStr += "Printing Track filename and Blob number\n"
        printStr += "---------------------------------------\n"
        if len(self.trackDictionary) > 0:
            for key, value in self.trackDictionary.items():
                printStr += (str(key) + " ----> " + str(value) + "\n")
        else:
            printStr += "No blobs linked"

        printStr += "---------------------------------------\n"
        printStr += "Printing daughter tracks:\n"
        if len(self.daughtersDictionary) > 0:
            for key, value in self.daughtersDictionary.items():
                printStr += (str(key) + " ----> " + str(value) + "\n")
        else:
            printStr += "No daughters linked\n"
        
        printStr += "--------------------------------------\n"

        printStr += "Areas:\n"
        if len(self.areas) > 0:
            for key, value in self.areas.items():
                printStr += (str(key) + '---->' + str(value) + "\n")
        else:
            printStr += "No areas found\n"
        printStr += "-------------------------------------\n"

        printStr += "Species: " + str(self.species) + "\n"

        printStr += "Track Length: " + str(self.trackLength) + "\n"

        printStr += "Fluor Channels: " + str(self.fluorChannels) + "\n"
        
        return printStr

    def __len__(self):
        return self.trackLength
    

    def add(self, fileNumber, blobNumber):
        fileName = self.dirName + str(fileNumber) + self.fileformat
        self.trackDictionary[fileName] = blobNumber
        self.fileBlobsDict[fileNumber] = blobNumber
        self.props[fileName] = regionprops(label(imread(fileName)))[int(blobNumber)]
        self.areas[fileName] = self.props[fileName]['area']
        self.lengths[fileName] = self.props[fileName]['major_axis_length']
        self.trackLength += 1

    def setDaughter(self, fileNumber, blobNumber):
        fileName = self.dirName + str(fileNumber) + self.fileformat
        if fileName in self.daughtersDictionary:
            self.daughtersDictionary[fileName].append(blobNumber)
        else:
            self.daughtersDictionary[fileName] = [blobNumber]
    

    def setParent(self, fileName, blobNumber):
        self.parent[fileName] = blobNumber

    def numDaughters(self):
        return len(self.daughtersDictionary)

    def numParents(self):
        return len(self.parent)




    # Add times form the metadata file somehow, to keep track 
    # of actual times
    def addTimes(self):
        pass

    def setSpecies(self, species):
        self.species = species

    def plotAreas(self):
        areasPlot = []
        for fileName, area in sorted(self.areas.items(), key= lambda kv : int(kv[0].split('/')[-1].split('.')[0])):
            areasPlot.append(area)
        
        # you can add time points loop here later
        plt.figure()
        plt.plot(areasPlot, 'r-')
        plt.show()

    def averageArea(self):
        areasPlot = []
        for fileName, area in sorted(self.areas.items(), key= lambda kv : int(kv[0].split('/')[-1].split('.')[0])):
            areasPlot.append(area)
        return np.mean(np.asarray(areasPlot))
 
    
    def calculateGrowth(self):
        sortedAreasDict = sorted(self.areas.items(), key= lambda kv : int(kv[0].split('/')[-1].split('.')[0]))
        startTime = int(sortedAreasDict[0][0].split('/')[-1].split('.')[0])
        endTime = int(sortedAreasDict[-1][0].split('/')[-1].split('.')[0])
        time = startTime
        for filename, areas in sortedAreasDict:
            self.areas_np[time] = areas
            time += 1

        for i in range(startTime+1, endTime+1):
            self.growth_np[i] = (self.areas_np[i] - self.areas_np[i-1])/self.areas_np[i-1]

        # Add the daughter calculation here
        # TODO
        
        return self.growth_np

    # Used for plotting the growth of one cell track
    def plotGrowth(self):
        pass

    # Used for splitting the track if it only has one daughter
    def split(self):
        pass


    def rollingGrowthRate(self, width = 5, fitatleast = 3):
        sortedAreasDict = sorted(self.areas.items(), key= lambda kv : int(kv[0].split('/')[-1].split('.')[0]))
        startTime = int(sortedAreasDict[0][0].split('/')[-1].split('.')[0])
        endTime = int(sortedAreasDict[-1][0].split('/')[-1].split('.')[0])
        time = startTime
        for filename, areas in sortedAreasDict:
            self.areas_rolling_np[time] = areas
            time += 1
        #print(self.areas_rolling_np)
        
        for i in range(width-1, len(self.areas_rolling_np)):
            current_areas = self.areas_rolling_np[i - width + 1 :i + 1]
            current_times = np.arange(i - width + 1, i + 1)
            set_areas = np.sum(current_areas != -1)
            #print(f"{current_times} ---> {current_areas} ---> {set_areas}")
            # you would need 3 points atleast to fit an exponential curve, you can filter for more points
            # by changing the width and fitatleast parameters
            # to break out rigth after you leave the track use
            #if set_areas >= fitatleast and current_areas[-1] != -1
            if set_areas >= fitatleast : 
                valid_areas_indices = np.where(current_areas != -1)
                valid_areas = current_areas[valid_areas_indices]
                valid_times = current_times[valid_areas_indices]
                #print(f"Valid: {valid_times} ----> {valid_areas}")
                popt, _ = curve_fit(exp_growth_fit, valid_times - valid_times[0], valid_areas)
                #print(f"Growth Rate {-1 * popt[1]}")
                self.growth_rolling_np[i] = -1 * popt[1]

        return self.growth_rolling_np

    def plotRollingGrowth(self, width = 5, fitatleast = 3, colorMap = colorMap, ylim = (-0.04, 0.06)):
        growth_rates = self.rollingGrowthRate(width = width, fitatleast = fitatleast)
        existing_growth_rates = growth_rates[np.where(growth_rates != -1)]
        time_points = np.where(growth_rates != -1)[0]
        
        plt.figure()
        plt.plot(time_points, existing_growth_rates, str(colorMap[self.species]) + '*')
        plt.xlim(0, self.exptTime)
        plt.ylim(*ylim)
        plt.show()

    def getWritableData(self):
        # Pacakge the useful data from the track in a correct format, for compression 
        # and saving and reloading.
        return




class singleChannelTrackingSiamese(Dataset):

    """
    Arguments:
         dirName -- dirname of the segmentation of individual channels
         net -- Net that is loaded in eval mode, will use no_grad() and pass blobs

    """
    def __init__(self, dirName, trackingParameters = None,backgroundFishData=None, net=None, fileformat='.tiff',
            maxCentroidMovement=50, channelSize=(150, 850), fishTransforms='box',
            fluorChThreshold=None, flourStdThreshold = 1.0, minFlourBoxHeight=90):

        self.dirName = dirName
        self.fileformat = fileformat
        self.siameseNet = net
        self.imgStdSize = (96, 32)
        self.maxCentroidMovement = maxCentroidMovement
        self.threshold = trackingParameters['threshold']
        self.differenceThreshold = trackingParameters['differenceThreshold']
        self.lastBlobCentroidCutoff = trackingParameters['lastBlobCentroidCutoff']
        self.divisionAreaRatio = trackingParameters['divisionAreaRatio']
        self.maxAreaGrowth = trackingParameters['maxAreaGrowth']
        self.exptTime = trackingParameters['exptTime']
        self.speciesMap = trackingParameters['speciesMap']
        self.numFluors = trackingParameters['numFluors']
        self.doneTracking = False

        self.transform = singleBlobTensorize()

        self.fileindices = [int(filename.split('.')[0].split('/')[-1]) for filename in 
                        glob.glob(self.dirName + "*" + self.fileformat)]

        self.fileindices.sort()
        self.nImages = len(self.fileindices)

        self.filenames = []
        self.images = []
        self.properties = []
        self.nBlobsPerImage = []

        for i in range(self.nImages):
            filename = self.dirName + str(self.fileindices[i]) + self.fileformat
            image = imread(filename) 
            props = regionprops(label(image))
            self.filenames.append(filename)
            self.nBlobsPerImage.append(len(props))
            self.images.append(image)
            self.properties.append(props)
        
        # data for net
        self.dataForNet = []

        # set fish data
        self.fishDirectory = self.dirName.split('blobs')[0] + 'fishChannels' + self.dirName.split('blobs')[-1]
        self.fishdata = singleChannelFishData(self.fishDirectory, backgroundFishData.channelNames,
                    channelSize=channelSize, transforms=fishTransforms, fluorChThreshold=fluorChThreshold,
                    minboxHeight=minFlourBoxHeight, backgroundFishData=backgroundFishData)
        
        if self.fishdata.getNumberFluorChannels() <= self.numFluors:
            self.doneTracking = True
            self.createDataForNet()

        self.speciesNames = None

    
    def createDataForNet(self):
        for fileIndex in range(len(self.filenames) - 1):
            prevfilename = self.filenames[fileIndex]
            currfilename = self.filenames[fileIndex + 1]
            # No loop throught each filename and 
            # get the blobs from previous and current filename and do some transforms
            # and then add the bundles to the dataset
            for blob1index in range(self.nBlobsPerImage[fileIndex]):
                blob1 = self.transform(self.getBlobForNet(fileIndex, blob1index))
                for blob2index in range(self.nBlobsPerImage[fileIndex + 1]):
                    blob2 = self.transform(self.getBlobForNet(fileIndex + 1, blob2index))

                    #if abs(blob1['props'][5].item() - blob2['props'][5].item()) < self.maxCentroidMovement:
                    self.dataForNet.append((blob1['props'], blob1['image'], blob2['props'], blob2['image'],
                                prevfilename, blob1index, currfilename, blob2index))
    
    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        
        return self.dataForNet[idx]

    def __len__(self):
        return len(self.dataForNet)


    def getBlobForNet(self, fileIndex, blobIndex):
        blobProperties = []
        blobProperties.extend(list(self.properties[fileIndex][blobIndex]['bbox']))
        blobProperties.append(self.properties[fileIndex][blobIndex]['area'])
        blobProperties.extend(list(self.properties[fileIndex][blobIndex]['centroid']))
        blobProperties.append(self.properties[fileIndex][blobIndex]['eccentricity'])
        blobProperties.append(self.properties[fileIndex][blobIndex]['major_axis_length'])
        blobProperties.append(self.properties[fileIndex][blobIndex]['minor_axis_length'])
        blobProperties.append(self.properties[fileIndex][blobIndex]['perimeter'])
        image = self.properties[fileIndex][blobIndex]['image']
        
        # Do some operations on the image to center the image in the field,
        # to standardize the inputs to the network
        height, width = image.shape
        #print("Image shape before:", image.shape)
        # set blank background image
        height_outside_limit = height > self.imgStdSize[0]
        width_outside_limit = width > self.imgStdSize[1]

        if height_outside_limit:
            image = resize(image, (self.imgStdSize[0], width))
        if width_outside_limit:
            image = resize(image, (height, self.imgStdSize[1]))

        background = np.zeros(self.imgStdSize)

        # recalculate height after rescaling in case
        height, width = image.shape
        if height%2 == 1:
            xlim_1 = self.imgStdSize[0]//2 - height//2
            xlim_2 = self.imgStdSize[0]//2 + height//2 + 1
        elif height%2 == 0:
            xlim_1 = self.imgStdSize[0]//2 - height//2
            xlim_2 = self.imgStdSize[0]//2 + height//2

        if width%2 == 1:
            ylim_1 = self.imgStdSize[1]//2 - width//2
            ylim_2 = self.imgStdSize[1]//2 + width//2 + 1
        elif width%2 == 0:
            ylim_1 = self.imgStdSize[1]//2 - width//2
            ylim_2 = self.imgStdSize[1]//2 + width//2

        #print(xlim_1, xlim_2, ylim_1, ylim_2)
        #print("Image shape:", image.shape)
        
        background[xlim_1 : xlim_2, ylim_1: ylim_2] = image
        return {'props': blobProperties, 'image': background}



    def getFileNumber(self, filename):
        return int(filename.split('.')[0].split('/')[-1])
    # From the net you get scores as output, you will have to construct appropriate links 
    # here before 
    def constructLinks(self, prevfilenames, currfilenames, prevBlobNumbers, currBlobNumbers, netScores):
        self.fileNamesProcessed = []
        self.links = []
        self.scoresArray = np.zeros(shape=(len(prevfilenames), 5))
        for i in range(len(prevfilenames)):
            self.scoresArray[i, 0] = self.getFileNumber(prevfilenames[i]) 
            self.scoresArray[i, 1] = self.getFileNumber(currfilenames[i]) 
            self.scoresArray[i, 2] = prevBlobNumbers[i]
            self.scoresArray[i, 3] = currBlobNumbers[i]
            self.scoresArray[i, 4] = netScores[i][0]

        for fileIndex in range(len(self.filenames) - 1):

            alreadyLinked = [] # trying to avoid linking things already linked

            for blob1index in range(self.nBlobsPerImage[fileIndex]):
                scores = -np.zeros((self.nBlobsPerImage[fileIndex + 1]))
                for blob2index in range(self.nBlobsPerImage[fileIndex + 1]):
                    scores[blob2index] = self.getScore(fileIndex, fileIndex + 1, blob1index, blob2index)
                
                #indicesAfterCutoff = np.where(scores < self.threshold)[0]
                sorted_scores = np.sort(scores)

                #if len(indicesAfterCutoff) == 1 and (indicesAfterCutoff[0] not in alreadLinkedInNext):
                #    self.links.append([fileIndex, fileIndex+1, blob1index, indicesAfterCutoff[0]])
                #   alreadLinkedInNext.append(indicesAfterCutoff[0])
                
                #elif len(indicesAfterCutoff) == 2:
                    # check if both pass the difference threshold
                    #daughter1Index = indicesAfterCutoff[0]
                    #daughter2Index = indicesAfterCutoff[1]

                    #if (abs(scores[daughter1Index] - scores[daughter2Index]) <= self.differenceThreshold):
                    #    # link both if area constraints are met
                    #    if ((self.properties[fileIndex+1][daughter1Index]['area'] < self.divisionAreaRatio * self.properties[fileIndex][blob1index]['area'])
                    #        and (self.properties[fileIndex+1][daughter2Index]['area'] < self.divisionAreaRatio * self.properties[fileIndex][blob1index]['area'])):
                    #        if (daughter1Index not in alreadLinkedInNext):
                    #            self.links.append([fileIndex, fileIndex+1, blob1index, daughter1Index])
                    #            alreadLinkedInNext.append(daughter1Index)
                    #        if (daughter2Index not in alreadLinkedInNext):
                    #            self.links.append([fileIndex, fileIndex + 1, blob1index, daughter2Index])
                    #            alreadLinkedInNext.append(daughter2Index)
                    #    else:
                    #        linkedblobIndex = daughter1Index
                    #        if daughter1Index not in alreadLinkedInNext:
                    #            self.links.append([fileIndex, fileIndex+1, blob1index, linkedblobIndex])
                    #            alreadLinkedInNext.append(linkedblobIndex)
                    #        elif (linkedblobIndex in alreadLinkedInNext):
                    #            linkedblobIndex = daughter2Index
                    #            self.links.append([fileIndex, fileIndex+1, blob1index, linkedblobIndex])
                    #            alreadLinkedInNext.append(linkedblobIndex)
           
                if len(sorted_scores) == 1:
                # check if it is less than threshold and add up to the links
                # and also the centroid of teh blob2 is not past the cutoff
                    if(sorted_scores <= self.threshold and
                    self.properties[fileIndex+1][blob2index]['centroid'][0] < self.lastBlobCentroidCutoff):
                        self.links.append([fileIndex, fileIndex+1, blob1index, blob2index])
                        alreadyLinked.append(blob2index)
                    
                elif len(sorted_scores) >= 2:
                    # check the difference between the first elements and if they pass the threshold
                    # find the indiecs of thes two elements as they give the blob index

                    if((sorted_scores[0] <= self.threshold) and (sorted_scores[1] <= self.threshold)):
                        # both are less than threshold, check the difference and add links
                        # They are daughters, find the blob indices in the original array scores and
                        # add them to the links
                        daughterIndex1 = np.where(scores == sorted_scores[0])[0][0]
                        daughterIndex2 = np.where(scores == sorted_scores[1])[0][0]
                        if(abs(sorted_scores[0]  - sorted_scores[1]) <= self.differenceThreshold):
                            # check for area constrains and alrad linked or not constraint
                            if((self.properties[fileIndex+1][daughterIndex1]['area'] < self.divisionAreaRatio * self.properties[fileIndex][blob1index]['area'])
                            and (self.properties[fileIndex+1][daughterIndex2]['area'] < self.divisionAreaRatio * self.properties[fileIndex][blob1index]['area'])):
                                if (daughterIndex1 not in alreadyLinked):
                                    self.links.append([fileIndex, fileIndex + 1, blob1index, daughterIndex1])
                                    alreadyLinked.append(daughterIndex1)
                                if (daughterIndex2 not in alreadyLinked):
                                    self.links.append([fileIndex, fileIndex + 1, blob1index, daughterIndex2])
                                    alreadyLinked.append(daughterIndex2)
                            else:
                                linkedblobIndex = daughterIndex1
                                if(linkedblobIndex not in alreadyLinked):
                                    self.links.append([fileIndex, fileIndex+1, blob1index, linkedblobIndex])
                                    alreadyLinked.append(linkedblobIndex)
                                elif(linkedblobIndex in alreadyLinked):
                                    linkedblobIndex = daughterIndex2
                                    self.links.append([fileIndex , fileIndex+1, blob1index, linkedblobIndex])
                                    alreadyLinked.append(linkedblobIndex)
                        
                        else:
                            # They don't reach the differene threshold
                            # so only link the closed blob, if it is not linked already, if linked, go
                            # to the next blob
                            linkedblobIndex = daughterIndex1
                            if (linkedblobIndex not in alreadyLinked):
                                self.links.append([fileIndex, fileIndex +1, blob1index, linkedblobIndex])
                                alreadyLinked.append(linkedblobIndex)
                            elif (linkedblobIndex in alreadyLinked):
                                linkedblobIndex = daughterIndex2
                                self.links.append([fileIndex, fileIndex + 1, blob1index, linkedblobIndex])
                                alreadyLinked.append(linkedblobIndex)
                    
                    # only one score is less than threshold
                    elif (sorted_scores[0] <= self.threshold and sorted_scores[1] > self.threshold):
                        # Just the least scores is less than threshold
                        linkedblobIndex = np.where(scores == sorted_scores[0])[0][0]
                        if (linkedblobIndex not in alreadyLinked):
                            if(self.properties[fileIndex+1][linkedblobIndex]['centroid'][0] < self.lastBlobCentroidCutoff):
                                self.links.append([fileIndex, fileIndex+1, blob1index, linkedblobIndex])
                                alreadyLinked.append(linkedblobIndex)
                    
                    
                
            self.fileNamesProcessed.append(self.dirName + str(fileIndex) + self.fileformat)

    # Look in side the lines of score structure and spit out the float value of the
    # score with the currentFileaname, nextFileName, blob1index, blob2index
    def getScore(self, currentFileIndex, nextFileIndex, blob1index, blob2index):
        #linksArray = np.asarray(self.scoresStructure)
        indices = np.where((self.scoresArray[:, :4] == [currentFileIndex, nextFileIndex, blob1index, blob2index]).all(axis = 1))[0]
        if len(indices) == 0:
            return -1.0 #if no match return a large value instead
        else:
            return float(self.scoresArray[indices][0][4])
                

    # convert the links made from scores, to convert to track structures
    def convertLinksToTracks(self):

        self.tracks = []
        # Here you will have to do the clean up as well
        
        # In this function you will have to construct the lineages of one cell and then
        # add it to the list of tracks

        # TODO: Here is where you construct tracks

        # TODO: Needs a bit more sophistication depending on how good 
        # the track links are
        # This is a function that constructs tracks from the links which 
        # is a array of list: [file1, file2, blob1, blob2].
        # Iterate till you go no where or have two blobs connected to one blob
        # 
        #self.tracks = [] # a list of dictionaries, each track is a dictionary
        # keys are frame numbers and values are indices of the blobs
        # create local copies tot not change the classes internal ones
        linksArray = np.asarray(self.links)
        
        while linksArray.size != 0:
        #for row in range(linksArray.shape[0]):
            # start with one element and end when the track ends
            # or when there are two blobs connecting the same blob
            first_row = linksArray[0] # start linking from here
            one_track = oneCellLineage(self.dirName, exptTime= self.exptTime)
            one_track.add(first_row[0], first_row[2]) # add the first blob
            # Now loop till there is nothing matching
            # How many blobs is this blob connected to : this blob means the first row
            while True:
                current_blob_index = np.where((linksArray[:, 0::2] == [first_row[0], first_row[2]]).all(axis=1))[0]
                #print("Current_blob_index: ", current_blob_index)
                connections_from_current_blob = np.where((linksArray[:, 0::2] == [first_row[0], first_row[2]]).all(axis=1))[0]
                if(len(connections_from_current_blob) == 2):
                    # Time to split the track and not add the track, find out which are daughters and keep track
                    #print("Blob has daughter and split: ", first_row, connections_from_current_blob)
                    daughterIndex1 = connections_from_current_blob[0]
                    daughterIndex2 = connections_from_current_blob[1]
                    one_track.setDaughter(linksArray[daughterIndex1][1], linksArray[daughterIndex1][3])
                    #print(linksArray[daughterIndex1][1], linksArray[daughterIndex1][3])
                    #print(linksArray[daughterIndex2][1], linksArray[daughterIndex2][3])
                    one_track.setDaughter(linksArray[daughterIndex2][1], linksArray[daughterIndex2][3])
                    linksArray = np.delete(linksArray, (current_blob_index), axis = 0)
                    break
                elif(len(connections_from_current_blob) == 1):
                    area_ratio = self.properties[first_row[1]][int(first_row[3])]['area']/self.properties[first_row[0]][int(first_row[2])]['area']
                    #print(area_ratio)
                    # link only if the area ratio falls in certain range
                    if area_ratio > 0.7 and area_ratio < (1/0.7):
                        one_track.add(first_row[1], first_row[3])
                        linksArray = np.delete(linksArray, (current_blob_index), axis = 0)
                        next_blob_index = np.where((linksArray[:,0::2] == [first_row[1], first_row[3]]).all(axis=1))[0] # Grab the index of the next blob from the array
                        if(next_blob_index.size != 0):
                            first_row = linksArray[next_blob_index[0]]
                    # one daughter case
                    elif area_ratio < 0.7 and area_ratio > 0.35:
                        daughterIndex1 = connections_from_current_blob[0]
                        one_track.setDaughter(linksArray[daughterIndex1][1], linksArray[daughterIndex1][3])
                        linksArray = np.delete(linksArray, (current_blob_index), axis = 0)
                        break
                    # don't link anything if area ration is off
                    else:
                        linksArray = np.delete(linksArray, (current_blob_index), axis = 0)
                        break

                elif(len(connections_from_current_blob) == 0):
                    break
            #print(one_track)
            self.tracks.append(one_track)
        



    def labelTracksWithFluorChannels(self, maxIterations=10, printFluorLabels=False):
        if self.fishdata == None:
            print("Fish linking failed")
            return None

        #print(f"Labeling some of the {len(self.tracks)}")

        # label all thracks that end in the last frame first
        lastFileName = self.fileNamesProcessed[-1]
        for cellLineage in self.tracks:
            # check if the track ends in the lastFrame
            if lastFileName in cellLineage.trackDictionary:
                # label track with correct flour channels to each of the track
                centroid = cellLineage.props[lastFileName]['centroid']
                cellLineage.fluorChannels = self.getFluorChannels(centroid)
                #print(cellLineage.fluorChannels)
        
        # Loop over and set the internal daughter indices of each track in the set of tracks
        self.setDaughterIndices()
        #print("Set internal daughter indices")
        while(maxIterations > 0):
            for _, oneLineage in enumerate(self.tracks, 0):
                indices = oneLineage._indexToDaughters
                #print(f"{indices} ----> {oneLineage.fluorChannels}")
                if oneLineage.fluorChannels == None and len(indices) > 0:
                    # set it to one of the daughters
                    for index in indices:
                        oneLineage.fluorChannels = self.tracks[index].fluorChannels
                        if oneLineage.fluorChannels != None:
                            break
            maxIterations -= 1

        # Loop over and check if the tracks's and their daughter species mathc
        # If not label it as conflict

        for _, oneLineage in enumerate(self.tracks, 0):
            daughterIndices = oneLineage._indexToDaughters
            for index in indices:
                if oneLineage.fluorChannels != self.tracks[index].fluorChannels:
                    oneLineage.fluorChannels = ['Conflict']
                    break
        if printFluorLabels == True:
            for oneLineage in self.tracks:
                print(oneLineage.fluorChannels)
    

    # Loop over the set of tracks and set the internal indices of the lineages in the overall
    # tracks list
    def setDaughterIndices(self):

        for index, oneLineage in enumerate(self.tracks, 0):
            oneLineage._indexToDaughters = []
            if oneLineage.numDaughters() > 0:
                # get daughter track's index in the set of tracks
                
                daughterStartFileName = list(oneLineage.daughtersDictionary.keys())[0]
                daughterBlobIndices = oneLineage.daughtersDictionary[daughterStartFileName]
                #print(f"{daughterStartFileName} ---> {daughterBlobIndices}")

                # Now loop over and set internal indices of the single track variables
                for index1, oneLineage1 in enumerate(self.tracks, 0):
                    if daughterStartFileName in oneLineage1.trackDictionary and oneLineage1.trackDictionary[daughterStartFileName] in daughterBlobIndices:
                        oneLineage._indexToDaughters.append(index1)

    # Function that takes in a species map and goes through all the tracks and 
    # assigns species labels.
    # Species Map is a mapping from the fluorescence channels to the name of the species
    def setSpeciesForAllTracks(self):
        self.speciesNames = []
        for species in self.speciesMap:
            self.speciesNames.append(species)
        
        for oneLineage in self.tracks:
            if oneLineage.fluorChannels != None:
                for species, fluorChannels in self.speciesMap.items():
                    if fluorChannels == set(oneLineage.fluorChannels):
                        #set species
                        oneLineage.species = species
                        #print(oneLineage.species)
                        break


    def insideBox(self, boxes, centroid):
        # Checking only x coordinate and forgetting about y coordinate of the boxes for now.
        # TODO: check y corrdinate later # might be useful when you have errors in chip making
        for box in boxes:
            if centroid[0] >= box[0][1] and centroid[0] <= box[0][1] + box[2]:
                return True

        return False

    def getFluorChannels(self, centroid):
        # loop over the fluor channels and return a list of all the channels in which
        # the centroid lies
        fluorChannels = []
        for channel in self.fishdata.channelNames:
            if self.insideBox(self.fishdata.fishBoxes[channel], centroid):
                fluorChannels.append(channel)
        
        if len(fluorChannels) == 0:
            return None
        else:
            return fluorChannels
        

    def calculateGrowth(self, minTrackLength = 5, minArea=100):
        channelGrowth = []
        tracks = self.convertLinksToTracks()
        for track in tracks:
            if len(track) > minTrackLength and track.averageArea() > minArea:
                channelGrowth.append(track.calculateGrowth())
        
        growthRates = np.asarray(channelGrowth)
        ntracks, ntimepoints = growthRates.shape
        mean_growth = np.zeros((ntimepoints,))
        for timepoint in range(ntimepoints):
            growth_t = growthRates[:, timepoint]
            set_growth = growth_t[np.where(growth_t != -1)]
            if set_growth.size != 0:
                mean_growth[timepoint] = np.mean(set_growth)

        return mean_growth, growthRates


    def getGrowthOfSpeciesLabeledTracks(self, width=5, fitatleast=3):
        growthRates = {}
        for species in self.speciesNames:
            growthRates[species] = []

        for i, track in enumerate(self.tracks, 0):
            if track.species != None:
                growth = track.rollingGrowthRate(width = width, fitatleast = fitatleast)
                growthRates[track.species].append(growth)
                #print(f"Track number --> {i}")
        return growthRates


    # TODO: Add more constraints and filters if needed like you did in 
    # calculateGrowth function like minAverage are and min Track Length (This will be
    # taken care of in the rolling case by other parameters like width and fitatleast)
    def calculateRollingGrowthRates(self, width=5, fitatleast=3):
        channelGrowth = dict()
        for species in self.speciesNames:
            channelGrowth[species] = []
        
        for track in self.tracks:
            if track.species != None:
                channelGrowth[track.species].append(track.rollingGrowthRate(width=width, fitatleast=fitalteast))

        for species in channelGrowth:
            channelGrowth[species] = np.asarray(channelGrowth[species])
        
        return channelGrowth


    def plotAllTracks(self, spacing=0):
        # Call this function only after calling trackFullDir() --> can fix later 
        # to run in loop dynamically
        # Function plots all the track links read from links list
        unlabeleddata = singleChannelDataset(self.dirName, fileformat=self.fileFormat)
        
        props = [] 
        for i in range(len(unlabeleddata)):
            regionpropsOneImage = regionprops(label(unlabeleddata[i]))
            props.append(regionpropsOneImage)

        n_images = len(unlabeleddata)
        print("Number of images: ", n_images)
        if n_images > 0: 
            height, width = unlabeleddata[0].shape
            print("Height: ", height, " Width: ", width)
        full_img = np.zeros(shape=(height, n_images * (width + spacing)))
        #indices = unlabeleddata.indices
        print("Full image shape: ", full_img.shape)

        # set the images
        for i in range(n_images):
            sub_image = unlabeleddata[i]
            spacing_image = np.zeros(shape=(height, spacing))
            sub_image = np.append(sub_image, spacing_image, axis = 1)
            full_img[:, i*(width + spacing) : (i+1) * (width+spacing)] = sub_image

        plt.figure()
        plt.imshow(full_img)

        # From the self.links generated after
        # 

        for row in self.links:
            # get frame_t from links
            frame_t = int(row[0].split('.')[0].split('/')[-1])
            blob_t, blob_t1 = row[2], row[3]
            #print(row)
            centroid_t_x, centroid_t_y = props[frame_t][blob_t]['centroid']
            centroid_t1_x, centroid_t1_y = props[frame_t + 1][blob_t1]['centroid']
            plt.plot([centroid_t_y + frame_t*(width + spacing), centroid_t1_y + (frame_t + 1)*(width + spacing)], [centroid_t_x, centroid_t1_x], 'r')

        plt.title('Links from tracking algorithm')
        plt.show(block = False)
    
    def getLinksBetweenTwoFrames(self, frame_i, frame_j):
        links = []
        for row in self.links:
            if self.getFrameNumber(row[0]) == frame_i and self.getFrameNumber(row[1]) == frame_j:
                links.append(row)

        return links
    
    def getFrameNumber(self, filename):
        return int(filename.split('.')[0].split('/')[-1])


    def plotAllTracksWithFISH(self, spacing=0):
               # Call this function only after calling trackFullDir() --> can fix later 
        # to run in loop dynamically
        # Function plots all the track links read from links list
        if self.fishdata == None:
            print("Fish data is not added")
            return
        
        #print("Number of images: ", n_images)

        if self.nImages > 0: 
            height, width = self.images[0].shape
            #print("Height: ", height, " Width: ", width)

        full_img = np.zeros(shape=(height, self.nImages * (width + spacing)))
        #indices = unlabeleddata.indices
        #print("Full image shape: ", full_img.shape)

        # set the images
        for i in range(self.nImages):
            sub_image = self.images[i]
            spacing_image = np.zeros(shape=(height, spacing))
            sub_image = np.append(sub_image, spacing_image, axis = 1)
            full_img[:, i*(width + spacing) : (i+1) * (width+spacing)] = sub_image

        fig, ax = plt.subplots(nrows=1, ncols = len(self.fishdata) + 1, gridspec_kw={'width_ratios':[self.nImages] + [1 for _ in range(len(self.fishdata))]})
        ax[0].imshow(full_img)

        # From the self.links generated after calling trackFullDir
        # 
        for row in self.links:
            # get frame_t from links
            frame_t = row[0]
            blob_t, blob_t1 = row[2], row[3]
            centroid_t_x, centroid_t_y = self.properties[frame_t][blob_t]['centroid']
            centroid_t1_x, centroid_t1_y = self.properties[frame_t + 1][blob_t1]['centroid']
            ax[0].plot([centroid_t_y + frame_t*(width + spacing), centroid_t1_y + (frame_t + 1)*(width + spacing)], [centroid_t_x, centroid_t1_x], 'r')

        ax[0].set(title ='Links from Siamese tracking algorithm -- Remembe these are links - Not final tracks')

        for i, channel in enumerate(self.fishdata.channelNames, 1):
            ax[i].imshow(self.fishdata[channel], cmap='gray')
            if self.fishdata.transforms == 'box' and len(self.fishdata.fishBoxes[channel]) != 0:
                for box in self.fishdata.fishBoxes[channel]:
                    ax[i].add_patch(Rectangle(*box, linewidth = 1, edgecolor='r', facecolor ='none'))
            ax[i].set(title=channel)
        plt.show(block = False)
    

    def plotTracksUsedForGrowth(self, spacing=0,
                colorMap={'Klebsiella': 'r', 'E.coli': 'b', 'Pseudomonas': 'g', 'E.cocci' : 'm', 'NoTrack': 'c'}):

        if self.fishdata == None:
            print("Fish data is not added")

        if self.nImages > 0:
            height, width = self.images[0].shape

        full_img = np.zeros(shape=(height, self.nImages * (width + spacing)))
        for i in range(self.nImages):
            sub_image = self.images[i]
            spacing_image = np.zeros(shape=(height, spacing))
            sub_image = np.append(sub_image, spacing_image, axis = 1)
            full_img[:, i * (width + spacing) : (i + 1) * (width + spacing)] = sub_image

        #Plot all the images
        fig, ax = plt.subplots(nrows=1, ncols = len(self.fishdata) + 1, gridspec_kw={'width_ratios': [self.nImages] + [1 for _ in range(len(self.fishdata))]})
        ax[0].imshow(full_img, cmap='gray')

        # Plot all the fish channels data and boxes around the thresholded data 
        for i, channel in enumerate(self.fishdata.channelNames, 1):
            ax[i].imshow(self.fishdata[channel], cmap='gray')
            if self.fishdata.transforms == 'box' and len(self.fishdata.fishBoxes[channel]) != 0:
                for box in self.fishdata.fishBoxes[channel]:
                    ax[i].add_patch(Rectangle(*box, linewidth = 1, edgecolor='r', facecolor ='none'))
            ax[i].set(title=channel)
 
        # for each track in the set of tracks start plotting on the image
        for cellLineage in self.tracks:
            species = cellLineage.species
            if species == None:
                color = colorMap['NoTrack']
            else:
                color = colorMap[species]
            
            # start connecting centroids
            trackAndBlobFiles = list(cellLineage.fileBlobsDict)
            for i in range(len(trackAndBlobFiles) - 1):
                frame_t = trackAndBlobFiles[i]
                frame_t1 = trackAndBlobFiles[i+1]
                blob_frame_t = cellLineage.fileBlobsDict[frame_t]
                blob_frame_t1 = cellLineage.fileBlobsDict[frame_t1]
                centroid_t_x, centroid_t_y = self.properties[frame_t][blob_frame_t]['centroid']
                centroid_t1_x , centroid_t1_y = self.properties[frame_t1][blob_frame_t1]['centroid']
                ax[0].plot([centroid_t_y + frame_t*(width + spacing), centroid_t1_y + (frame_t + 1)*(width + spacing)], [centroid_t_x, centroid_t1_x], color)
        
        plt.show(block=False)



# File containing functions that will help in running the segmentation 
# functions on your dataset
from scipy import ndimage
import torch
import os
import time
import numpy as np
import multiprocessing as mp
from narsil.segmentation.network import basicUnet, smallerUnet
from narsil.segmentation.datasets import phaseFolder
from narsil.utils.transforms import resizeOneImage, tensorizeOneImage
from torchvision import transforms
from torch.utils.data import Dataset, DataLoader
from scipy.ndimage.morphology import binary_dilation, binary_erosion, binary_fill_holes, binary_opening
from scipy.signal import find_peaks
from skimage.io import imread, imsave
from skimage.transform import resize, rotate
from skimage.exposure import equalize_adapthist
from skimage.morphology import remove_small_holes, remove_small_objects
from skimage.measure import label, find_contours
from collections import OrderedDict
from functools import partial


############################################################
############################################################
############## Segmentation Functions ######################
######### (Cells and channel segmentaiton) #################
############################################################
############################################################

# Function to segment a directory conataining .tiff files
def segmentPosDirectory(positionPhaseDir, segCellNet, channelNet, segmentationParameters):
    """ Function to segment one directory containing images using the 
    parameters given in the segmentation parameters folder

    Arguments
    ---------
    positionPhaseDir: str
        Path of the directory containing phase contrast images that are to
        be segmented, string usually ends in position name ..../Pos101/ 
    segCellnet: torch model for cell segmentation
        Neural net loaded onto appropriate device and set to eval mode
	
	channelNet: torch model for channel segmentation
		Neural net loaded onto appropriate device and set to eval mode

    segmentationParameters: dict
        A dictionary containinig all the parameters needed for segmentation

    Returns
    -------
    Succcess/Failure: bool
        Success if everything goes well in this directory
    positionImgChop: dict
        A dictionary with keynames as imagefilenames and values are numpy arrays of the 
        locations where you can cut out the individual channels from the images
    """

    phaseImageTransform = segmentationParameters['transforms'] 

    device = torch.device(segmentationParameters['device'] if torch.cuda.is_available() else "cpu")

    positionNumber = int(positionPhaseDir.split('/')[-2][3:])
    #print(positionNumber)
    positionImgChopMap = {}

    if positionNumber in segmentationParameters['flipPositions']:
        flip = True
    else:
        flip = False
    
    phase_dataset = phaseFolder(positionPhaseDir, phaseDirName=segmentationParameters['phasePreset'],
                    transform=phaseImageTransform, phase_fileformat=segmentationParameters['fileformat'],
                    addNoise=segmentationParameters['addNoise'], flip=flip)

    phaseDataLoader = DataLoader(phase_dataset, segmentationParameters['batch_size'], shuffle=False, num_workers=6)

    channelCuttingFailed = False

    with torch.no_grad():
        for i_batch, data_test in enumerate(phaseDataLoader, 0):
            phase_test = data_test['phase'].to(device)
            filename_to_save = data_test['phase_filename'][0].split('.')[0].split('/')[-1]
            #print(filename_to_save)
            mask_pred = segCellNet(phase_test) 

            mask_pred = torch.sigmoid(mask_pred)

            # Grab the channel locations and write them to the dictionary
            if segmentationParameters['getChannelLocations'] and channelCuttingFailed == False:
                channel_pred = torch.sigmoid(channelNet(phase_test)) > segmentationParameters['channelSegThreshold']
                channel_pred = channel_pred.to("cpu").detach().numpy().squeeze(0).squeeze(0)
                #print(channel_pred.shape)
                try:
                    positionImgChopMap[filename_to_save] = getChannelLocationSingleImage(channel_pred, segmentationParameters['channelCuttingParameters'])
                
                except Exception as e:
                    channelCuttingFailed = True
                    print(e)
                    print(f"Pos{positionNumber} has failed cutting channels")
                    continue
            
            #print(f"{filename_to_save} ---> {channel_pred.shape}")
            # TODO: loop over if batch_size > 1
            mask_pred = mask_pred.to("cpu").detach().numpy().squeeze(0).squeeze(0)

            if segmentationParameters['useContourFinding']:
                mask_contour = np.zeros_like(mask_pred, dtype='bool')
                contours = find_contours(mask_pred, segmentationParameters['contoursThreshold'])
                for contour in contours:
                    mask_contour[np.round(contour[:, 0]).astype('int'), np.round(contour[:, 1]).astype('int')] = 1
                
                mask_contour = binary_fill_holes(mask_contour)
                mask_pred = mask_contour
            else:
                mask_pred = mask_pred > segmentationParameters['segmentationThreshold']

            if segmentationParameters['cleanSmallObjects']:
                mask_pred = remove_small_holes(mask_pred , area_threshold=30)
                mask_pred_labeled = label(mask_pred)
                mask_cleaned = remove_small_objects(mask_pred_labeled, min_size=40)
                mask_pred = mask_cleaned > 0

            if segmentationParameters['dilateAfterSeg']:
                mask_pred = binary_dilation(mask_pred)
            
            # clean up the regions and artifacts from the background
            if segmentationParameters['backgroundArtifactClean']:
                phase_original = data_test['phase_original'].numpy().squeeze(0).squeeze(0)
                #print(phase_original.shape)
                phase_original_thresholded1 = phase_original < segmentationParameters['minBackgroundOnPhase']
                phase_original_thresholded2 = phase_original > 0
                phase_original_thresholded = phase_original_thresholded1 * phase_original_thresholded2

                mask_pred = mask_pred * phase_original_thresholded

            if segmentationParameters['saveSeg']:
                segSaveDir = segmentationParameters['saveResultsDir'] + 'Pos' + str(positionNumber) + '/segmentedPhase/'
                if not os.path.exists(segSaveDir):
                    os.makedirs(segSaveDir)
                imsave(segSaveDir + filename_to_save + segmentationParameters['fileformat'], mask_pred.astype('float32'),
                    plugin='tifffile', compress=6, check_contrast=False)
            
            if segmentationParameters['savePhase']:
                phaseSaveDir = segmentationParameters['saveResultsDir'] + 'Pos' + str(positionNumber) + '/processedPhase/'
                phase_save = data_test['phase_original'].numpy().squeeze(0).squeeze(0)
                if not os.path.exists(phaseSaveDir):
                    os.makedirs(phaseSaveDir)
                imsave(phaseSaveDir + filename_to_save + segmentationParameters['fileformat'], phase_save.astype('float32'), plugin='tifffile', compress=6)
            
            if channelCuttingFailed == True:
                positionImgChopMap[filename_to_save] = np.asarray([])
        
        return (True, channelCuttingFailed, positionImgChopMap)

    # if doesn't return then something has failed during segmentation
    return (False, channelCuttingFailed, positionImgChopMap)


# Function to find channel locations to cut from the segmented channel image and some paramters
def getChannelLocationSingleImage(channel_pred, channelCuttingParameters):
    hist = np.sum(channel_pred[channelCuttingParameters['channel_min']: channelCuttingParameters['channel_max']], axis=0) > channelCuttingParameters['channel_sum']

    peaks, _ = find_peaks(hist, distance=channelCuttingParameters['histPeaksDistance'])

    indices_with_larger_gaps = np.where(np.ediff1d(peaks) > channelCuttingParameters['minBarcodeDistance'])[0]

    possible_barcode_locations = indices_with_larger_gaps[np.argmax(indices_with_larger_gaps > channelCuttingParameters['firstBarcodeIndex'])]

    #print(possible_barcode_locations)
    numChannels = channelCuttingParameters['numChannels']
    before_barcode_locations = np.zeros((numChannels,), dtype=int)
    after_barcode_locations = np.zeros((numChannels,), dtype=int)

    for i in range(numChannels):
        before_barcode_locations[i] = peaks[possible_barcode_locations-i]
        after_barcode_locations[i] = peaks[possible_barcode_locations+i+1]

    locations_to_cut = np.concatenate((before_barcode_locations[::-1], after_barcode_locations), axis = 0)

    return locations_to_cut



def loadNet(modelPath, device):
    """
    A function that takes in a model file and returns a net in eval mode 

    Arguments
    ---------
    modePath: .pth model file
        A model file, .pth file containing the model details to load the correct 
        model file

    device: str 
        usually "cuda:0" or "cuda:1"

    Returns
    -------
    net: torch.nn.Module object
        Net in eval mode
    """

    # Read the model file
    savedModel = torch.load(modelPath)

    # use the net depending on what model is loaded
    if savedModel['modelParameters']['netType'] == 'big':
        net = basicUnet(savedModel['modelParameters']['transposeConv'])
    elif savedModel['modelParameters']['netType'] == 'small':
        net = smallerUnet(savedModel['modelParameters']['transposeConv'])

    # load the net
    net.load_state_dict(savedModel['model_state_dict'])

    # send to device
    device = torch.device(device if torch.cuda.is_available() else "cpu")
    net.to(device)

    # eval mode
    net.eval()
    for p in net.parameters():
        p.requires_grad_(False)

    return net


def segmentAllPositions(phaseMainDir, positions, models, segmentationParameters):
    """ 
    Function to segment position ranges given to the function

    Arguments
    ---------
    phaseMainDir: str
        Path of the directory where positions with phase images live

    positions: range object or list 
        Range or list of positions in the directroy or the ones you want to 
        segment
    
    models: dict
        A dict with paths of different models (for segmentation of cells and 
            channels). "cells" and "channels" are keys and modelPaths are values

    segmentationParameters: dict
        Dict containing all the parameters needed for segmentation like GPU devices,
        image resize and shrink components, segmentation threshold, etc
    
    Returns
    -------
    None

    """
    start = time.time()

    device = segmentationParameters['device']

    net = loadNet(models['cells'], device)

    print("Segmentation Network loaded successfuly ...")


    channelNet  = loadNet(models['channels'], device)
    print("Channel Segmentation Network loaded successfully ... ")

    cuttingLocations = {}

    for position in positions:
        positionPhaseDir = phaseMainDir + 'Pos' + str(position) + '/'
        success, channelCuttingFailed, positionsImgChopMap = segmentPosDirectory(positionPhaseDir, net, channelNet, segmentationParameters)
        if (success and (channelCuttingFailed == False)):
            print(positionPhaseDir)
            cuttingLocations[position] = positionsImgChopMap

    duration = time.time() - start
    print(f"Duration for segmenting {positions} is {duration}s")

    print("Writing cutting locations ...")
    print(cuttingLocations.keys(), " positions have cutting locations done")
    np.save(segmentationParameters['saveResultsDir'] + 'channelLocations.npy', cuttingLocations)


def segmentDirectory(phaseDir, models, segmentationParameters, saveDir=None):
    """
    Function to segment a pure directory containing .tiff files. 
    You can use this to plot overlays as well.

    """
    return

################################################################
############ Cutting individual channel stacks #################
######################### & ####################################
############### Fluorescent channel Cutting ####################
################################################################

# cut inidividual channels from one position
def cutChannelsOnePosition(analysisPosDir, channelLocations, cuttingAndWritingParameters,
                 segmentedFileFormat='.tiff'):
    """
    Funciton to write and cut individual mother machine channels and write them for one
    position (which is a stack of images)
    """
    positionNumber = int(analysisPosDir.split('/')[-2][3:])

    # Directory where segmented images live, filenames are same as the raw data images
    segmentedPhaseDir = analysisPosDir + 'segmentedPhase/'
    
    blobsWriteDir = analysisPosDir + 'blobs/'
    channelWidth = cuttingAndWritingParameters['channelWidth']

    for i, filename in enumerate(channelLocations[positionNumber], 0):
        if (i > cuttingAndWritingParameters['cutUntilFrames']):
            return True
        
        imageFilename = segmentedPhaseDir + filename + segmentedFileFormat
        image = imread(imageFilename)

        peaks = channelLocations[positionNumber][filename]
        left = peaks - (channelWidth//2)
        right = peaks + (channelWidth//2)
        channelLimits = list(zip(left, right))

        for l in range(len(channelLimits)):
            if not os.path.exists(blobsWriteDir + str(l)):
                os.makedirs(blobsWriteDir + str(l))
            imgChop = image[:, channelLimits[l][0]: channelLimits[l][1]] * 255
            imsave(blobsWriteDir + str(l) + '/' + str(i) + segmentedFileFormat, imgChop, plugin='tifffile',
                         compress=6, check_contrast=False)
        
    return False



# Parallelize cutting channels and writing
def cutChannelsAllPositions(analysisMainDir, positions, cuttingAndWritingParameters,
        numProcesses=6):
    """
    Function to write and cut individual mother machine channels and write
    them to appropriate directories, for cell-tracking and growth rate analysis
    later

    Arguments
    ---------
    analysisMain: str
        Path of the analysis directory containing the segmented images directory

    positions: range object or list
        Range or list of position numbers to cut

    cuttingAndWritingParameters: dict
        Dictionary containing parameters used in channel cutting

    numProcesses: int
        Number of processes you want to parallelize on

    Returns
    -------
    None
    """
    start = time.time()

    channelLocationsFile = cuttingAndWritingParameters['saveResultsDir'] + 'channelLocations.npy'

    channelLocations = np.load(channelLocationsFile, allow_pickle=True).item()

    listPositionDirs = []
    for position in positions:
        if position not in channelLocations or len(channelLocations[position]) == 0:
            print(f"Skipping Pos{position} due to bad channel cutting ... ")
        else:
            listPositionDirs.append(analysisMainDir + 'Pos' + str(position) + '/')

    print(listPositionDirs)

    try:
        mp.set_start_method("spawn")
    except RuntimeError:
        pass
    
    pool = mp.Pool(processes=numProcesses)
    pool.map(partial(cutChannelsOnePosition, channelLocations=channelLocations, cuttingAndWritingParameters=cuttingAndWritingParameters), listPositionDirs)
    pool.close()
    pool.join()

    duration = time.time() - start
    print(f"Duration of cutting channels of {positions} is {duration}s")

    return None

###################################################################
################# Fluorescent cutting functions ###################
###################################################################

def cutFluorOnePosition(fluorPositionDir, channelLocations, fluorParameters):
    
    channelNames = fluorParameters['channelNames']
    positionNumber = int(fluorPositionDir.split('/')[-2][3:])
    channelWidth = fluorParameters['channelWidth']
    fluorTransform = fluorParameters['transform'] 

    writeDirectory = fluorParameters['saveResultsDir'] + 'Pos' + str(positionNumber) + '/fishChannels/'

    for channelName in channelNames:

        channelImageFileName = fluorPositionDir + channelName + '/' + fluorParameters['fluorImageName'] + '.tiff'
        channelImage = imread(channelImageFileName, as_gray=True)

        height, width = channelImage.shape
        if positionNumber in fluorParameters['flipPositions']:
            channelImage = rotate(channelImage, angle=180, preserve_range=True)

        # Here you can apply the transformations need to clean up the image
        channelImage = fluorTransform(channelImage)
        channelImage = channelImage.astype('uint16')
        if fluorParameters['equalize'] == 'equalize_adapthist':
            channelImage = (65535 * equalize_adapthist(channelImage))


        peaks = channelLocations[positionNumber][fluorParameters['phaseImageToMap']]
        left = peaks - (channelWidth//2)
        right = peaks + (channelWidth//2)
        channelLimits = list(zip(left, right))

        for l in range(len(channelLimits)):
            if not os.path.exists(writeDirectory + str(l)):
                os.makedirs(writeDirectory + str(l))
            imsave(writeDirectory + str(l) + '/' + channelName + '.tiff', channelImage[:, channelLimits[l][0] : channelLimits[l][1]], compress=6)

    print(f"FluorCutting in Pos{positionNumber} Done ..")
    return None

    

def cutFluorAllPositions(fluorMainDir, positions, fluorParameters, numProcesses=6):

    start = time.time()
    channelLocationsFilename = fluorParameters['saveResultsDir'] + 'channelLocations.npy'
    print(f"Reading {channelLocationsFilename} to get locations to cut ... ")

    channelLocations = np.load(channelLocationsFilename, allow_pickle=True).item()

    listPositionDirs = []
    for position in positions:
        if (position not in channelLocations) or len(channelLocations[position]) == 0:
            print(f"Skipping Pos{position} flour cutting due to bad channel detection ...")
        else:
            listPositionDirs.append(fluorMainDir + 'Pos' + str(position) + '/')
    print(listPositionDirs)

    try:
        mp.set_start_method("spawn")
    except RuntimeError:
        pass

    pool = mp.Pool(processes=numProcesses)
    pool.map(partial(cutFluorOnePosition, channelLocations=channelLocations, fluorParameters=fluorParameters), listPositionDirs)
    pool.close()
    pool.join()

    duration = time.time() - start
    print(f"Duration of cutting Fluorescent channels of {positions} is {duration}s")

    return None
 




from matplotlib.pyplot import fill
import numpy as np
import math
import random
import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision import transforms
import torchvision
import torchvision.transforms.functional as TF
from skimage import transform

class padTo16(object):

    def __call__(self, sample):
        
        width, height = sample['phase'].size
        pad_width = 16 - (width % 16) 
        pad_height = 16 - (height % 16)

        sample['phase'] = TF.pad(sample['phase'], padding=[0, 0, pad_width, pad_height], padding_mode="constant", fill=0)
        sample['mask'] = TF.pad(sample['mask'], padding=[0, 0, pad_width, pad_height], padding_mode="constant", fill=0)
        sample['weights'] = TF.pad(sample['weights'], padding=[0, 0, pad_width, pad_height], padding_mode="constant", fill=0)

        return sample


class changedtoPIL(object):

    def __call__(self, sample):
        phaseImg = sample['phase'].astype('int32')
        maskImg = sample['mask']
        weightsImg = sample['weights'].astype('float32')

        sample['phase'] = TF.to_pil_image(phaseImg)
        sample['mask'] = TF.to_pil_image(maskImg)
        sample['weights'] = TF.to_pil_image(weightsImg)
        
        return sample

class randomCrop(object):

    def __init__(self, output_size):
        self.output_size = output_size

    def __call__(self, sample):
        
        i, j, h, w = transforms.RandomCrop.get_params(sample['phase'], output_size=(self.output_size, self.output_size))

        sample['phase'] = TF.crop(sample['phase'], i, j, h, w)
        sample['mask'] = TF.crop(sample['mask'], i, j, h, w)
        sample['weights'] = TF.crop(sample['weights'], i, j, h, w)
        return sample

class randomVerticalFlip(object):

    def __call__(self, sample):

        if random.random() < 0.5:
            sample['phase'] = TF.vflip(sample['phase'])
            sample['mask'] = TF.vflip(sample['mask'])
            sample['weights'] = TF.vflip(sample['weights'])

        return sample

class randomRotation(object):

    def __init__(self, rotation_angle):
        self.rotation_angle = rotation_angle

    def __call__(self, sample):

        angle = transforms.RandomRotation.get_params(self.rotation_angle)
        sample['phase'] = TF.rotate(sample['phase'], angle)
        sample['mask'] = TF.rotate(sample['mask'], angle)
        sample['weights'] = TF.rotate(sample['weights'], angle)

        return sample

class randomAffine(object):

    def __init__(self, scale, shear):
        self.scale = scale # something like 0.75-1.25
        self.shear = shear # something like [-30, 30, -30, 30]

    def __call__(self, sample):

        angle, translations, scale, shear = transforms.RandomAffine.get_params(degrees=[0, 0], translate=None, 
                scale_ranges=self.scale, shears=self.shear, img_size=sample['phase'].size)

        sample['phase'] = TF.affine(sample['phase'], angle=angle, translate=translations, scale=scale, shear=shear)
        sample['mask'] = TF.affine(sample['mask'], angle=angle, translate=translations, scale=scale, shear=shear)
        sample['weights'] = TF.affine(sample['weights'], angle=angle, translate=translations, scale=scale, shear=shear)
        return sample

class randomBrightness(object):

    def __init__(self, probability):
        self.probability = probability

    def __call__(self, sample):
        
        if random.random() < self.probability:

            if random.random() < 0.5:
                brightness_factor = 1.0 + random.random()/5.0 # random 20% adjustment
                sample['phase'] = TF.adjust_brightness(sample['phase'], brightness_factor=brightness_factor)
            else:
                brightness_factor = 1.0 - random.random()/5.0 # random 20% adjustment
                sample['phase'] = TF.adjust_brightness(sample['phase'], brightness_factor=brightness_factor)

        return sample

class randomContrast(object):

    def __init__(self, contrast_factor, probability):
        self.contrast_factor = contrast_factor
        self.probability = probability
    
    def __call__(self, sample):
        
        if random.random() < self.probability:
            sample['phase'] = TF.adjust_contrast(sample['phase'], self.contrast_factor)

class toTensor(object):

    def __call__(self, sample):
        sample['phase'] = transforms.ToTensor()(np.array(sample['phase']).astype(np.float32))
        sample['mask'] = transforms.ToTensor()(np.array(sample['mask']).astype(np.float32))
        sample['weights'] = transforms.ToTensor()(np.array(sample['weights']).astype(np.float32))

        sample['phaseFilename'] = str(sample['phaseFilename'])
        sample['maskFilename'] = str(sample['maskFilename'])
        sample['weightsFilename'] = str(sample['weightsFilename'])
        return sample

class normalize(object):

    def __call__(self, sample):

        sample['phase'] = (sample['phase'] - torch.mean(sample['phase']))/torch.std(sample['phase'])

        # mask are 255 for True and 0 for false
        sample['mask'] = sample['mask']/255.0

        #sample['weights'] += 1.0
        

        return sample

class addnoise(object):

    def __init__(self, mean=0.0, std=0.15):
        self.mean = mean
        self.std = std

    def __call__(self, sample):

        sample['phase'] = sample['phase'] + torch.randn(sample['phase'].size()) * self.std + self.mean

        return sample


class resizeOneImage(object):

    def __init__(self, imgResizeShape, imgToNetSize):
        assert isinstance(imgResizeShape, tuple)
        assert isinstance(imgToNetSize, tuple)
        self.imgResizeShape = imgResizeShape
        self.imgToNetSize = imgToNetSize

    def __call__(self, image):

        height, width = image.shape
        # check if imageSize is bigger or equal to the 
        image = np.pad(image, pad_width=((0, self.imgResizeShape[0] - height), (0, self.imgResizeShape[1] - width)), 
                          mode='constant', constant_values = 0.0)
        if self.imgResizeShape[0] != self.imgToNetSize[0] or self.imgResizeShape[1] != self.imgToNetSize[1]:
            # Net size is not same a resized image
            image = transform.resize(image, self.imgToNetSize, anti_aliasing=True, preserve_range=True)
        return image


class tensorizeOneImage(object):
    def __init__(self, numUnsqueezes=1):
        self.numUnsqueezes = numUnsqueezes
    def __call__(self, phase_image):
        phase_image = phase_image.astype('float32')
        if self.numUnsqueezes == 1:
            return torch.from_numpy(phase_image).unsqueeze(0)
        elif self.numUnsqueezes == 2:
            return torch.from_numpy(phase_image).unsqueeze(0).unsqueeze(0)


class stripAxis(object):

    def __call__(self, imageTensor):
        return imageTensor.to("cpu").detach().numpy().squeeze(0).squeeze(0)


# Class that basically tensorizes the siamese databundles
class imgBundleTensorize(object):

    def __init__(self,  propertiesNormalization = None):
        self.propertiesNormalization = propertiesNormalization

    def __call__(self, siamBundle):
        imgBundle1 = siamBundle[0]
        imgBundle2 = siamBundle[1]
        linked = siamBundle[2] # True if linked, but we invert it so that it is 0 if they are same

        properties1 = imgBundle1['props']
        image1 = imgBundle1['image'].astype('float32')
        properties2 = imgBundle2['props']
        image2 = imgBundle2['image'].astype('float32')
        # convert properties and image to tensor
        if self.propertiesNormalization:
            properties1 = np.array(properties1).astype('float32') / self.propertiesNormalization
            properties2 = np.array(properties2).astype('float32') / self.propertiesNormalization
        else:
            properties1 = np.array(properties1).astype('float32') 
            properties2 = np.array(properties2).astype('float32') 
            
        return [{'props': torch.from_numpy(properties1),
                'image' : torch.from_numpy(image1).unsqueeze(0)}, {'props': torch.from_numpy(properties2),
                'image': torch.from_numpy(image2).unsqueeze(0)} , torch.from_numpy(np.array([int(not linked)], dtype=np.float32))]
     

class singleBlobTensorize(object):

    def __init__(self,  propertiesNormalization = None):
        self.propertiesNormalization = propertiesNormalization

    def __call__(self, blobBundle):

        properties = blobBundle['props']
        image = blobBundle['image'].astype('float32')
        # convert properties and image to tensor
        if self.propertiesNormalization:
            properties = np.array(properties).astype('float32') / self.propertiesNormalization
        else:
            properties = np.array(properties).astype('float32')
        
        
        return {'props': torch.from_numpy(properties).unsqueeze(0),
                'image' : torch.from_numpy(image).unsqueeze(0).unsqueeze(0)}




# File containing the operations that are done on fluorescent Images
# to obtain binary segmentation mask and weight masks
import numpy as np
from scipy import ndimage as ndi
from skimage.io import imread, imsave
import matplotlib.pyplot as plt
from skimage.measure import label, regionprops
from skimage.morphology import remove_small_objects
import glob
import pathlib
from skimage.restoration import rolling_ball
from skimage.filters import difference_of_gaussians, gaussian
from skimage.exposure import equalize_hist
from skimage.morphology import area_opening
from skimage.feature import peak_local_max
from skimage.segmentation import watershed
from skimage.util.dtype import img_as_ubyte
from skimage import img_as_ubyte

# Function to generate weight maps from a binary segmentation file. 
# The binary segmenation where dtype is uint8 with true values == 255
def generateWeights(filename, sigma = 5, w0 = 10):
	img = imread(filename)
	# removing objects and calculating distances to objects needs labelled images
	labeledImg, num = label(img, return_num=True, connectivity=2)
	# remove small objects
	labeledImg = remove_small_objects(labeledImg, min_size=250)
	# unique values == number of blobs
	unique_values = np.unique(labeledImg) 
	num_values = len(unique_values)
	h, w = labeledImg.shape
	# stack keeps distance maps each blob
	stack = np.zeros(shape=(num_values, h, w))
	for i in range(num_values):
		stack[i] = ndi.distance_transform_edt(~(labeledImg == unique_values[i]))
	# sort the distance
	sorted_distance_stack = np.sort(stack, axis=0)
	# d1 and d2 are the shortest and second shortest distances to each object, 
	# sorted_distance_stack[0] is distance to the background. One can ignore it
	distance_sum = sorted_distance_stack[1] + sorted_distance_stack[2]
	squared_distance = distance_sum ** 2/ (2 * (sigma**2))
	weightmap = w0 * np.exp(-squared_distance)*(labeledImg == 0)
	return weightmap

# Used for generating binary fluorescent mask from fluorescence data,
# Tested against Pseudomonas data only
def generateBinaryMaskFromFluor(fluorImgfilename):
	fluorImg = imread(fluorImgfilename).astype('float32')
	background = rolling_ball(fluorImg, radius = 50)
	fluorImg_remback = fluorImg - background
	filtered = difference_of_gaussians(fluorImg_remback, 1, 4)
	fluorImg_scaled_eq = equalize_hist(filtered)
	filtered_gaussian = gaussian(fluorImg_scaled_eq)
	filtered_median = ndi.median_filter(filtered_gaussian, size=5)
	filtered_gaussian  = filtered_gaussian > 0.92
	image_opening = area_opening(filtered_gaussian, area_threshold=150)
	distance = ndi.distance_transform_edt(image_opening)
	coords = peak_local_max(distance,min_distance = 90, footprint=np.ones((7,7)), labels=image_opening)
	mask = np.zeros(distance.shape, dtype=bool)
	mask[tuple(coords.T)] = True
	markers, _ = ndi.label(mask)
	labels = watershed(-distance, markers, mask=image_opening, watershed_line=True, connectivity=2)
	return img_as_ubyte(labels > 0)